#
# Copyright (c) 2018, Salesforce, Inc.
#                     The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import datetime
import json
import logging
import os
import subprocess

from .tasks.registry import get_tasks

logger = logging.getLogger(__name__)


def get_commit():
    directory = os.path.dirname(__file__)
    return subprocess.Popen("cd {} && git log | head -n 1".format(directory), shell=True,
                            stdout=subprocess.PIPE).stdout.read().split()[1].decode()


def save_args(args):
    os.makedirs(args.log_dir, exist_ok=args.exist_ok)
    with open(os.path.join(args.log_dir, 'config.json'), 'wt') as f:
        json.dump(vars(args), f, indent=2)


def parse_argv(parser):
    parser.add_argument('--root', default='.', type=str,
                        help='root directory for data, results, embeddings, code, etc.')
    parser.add_argument('--data', default='.data/', type=str, help='where to load data from.')
    parser.add_argument('--save', required=True, type=str, help='where to save results.')
    parser.add_argument('--embeddings', default='.embeddings', type=str, help='where to save embeddings.')
    parser.add_argument('--cache', default='.cache/', type=str, help='where to save cached files')

    parser.add_argument('--train_tasks', nargs='+', type=str, dest='train_task_names', help='tasks to use for training',
                        required=True)
    parser.add_argument('--train_iterations', nargs='+', type=int, help='number of iterations to focus on each task')
    parser.add_argument('--train_batch_tokens', nargs='+', default=[9000], type=int,
                        help='Number of tokens to use for dynamic batching, corresponding to tasks in train tasks')
    parser.add_argument('--jump_start', default=0, type=int, help='number of iterations to give jump started tasks')
    parser.add_argument('--n_jump_start', default=0, type=int, help='how many tasks to jump start (presented in order)')
    parser.add_argument('--num_print', default=15, type=int,
                        help='how many validation examples with greedy output to print to std out')

    parser.add_argument('--no_tensorboard', action='store_false', dest='tensorboard',
                        help='Turn off tensorboard logging')
    parser.add_argument('--tensorboard_dir', default=None,
                        help='Directory where to save Tensorboard logs (defaults to --save)')
    parser.add_argument('--max_to_keep', default=5, type=int, help='number of checkpoints to keep')
    parser.add_argument('--log_every', default=int(1e2), type=int, help='how often to log results in # of iterations')
    parser.add_argument('--save_every', default=int(1e3), type=int,
                        help='how often to save a checkpoint in # of iterations')

    parser.add_argument('--val_tasks', nargs='+', type=str, dest='val_task_names',
                        help='tasks to collect evaluation metrics for')
    parser.add_argument('--val_every', default=int(1e3), type=int,
                        help='how often to run validation in # of iterations')
    parser.add_argument('--val_no_filter', action='store_false', dest='val_filter',
                        help='whether to allow filtering on the validation sets')
    parser.add_argument('--val_batch_size', nargs='+', default=[256], type=int,
                        help='Batch size for validation corresponding to tasks in val tasks')

    parser.add_argument('--vocab_tasks', nargs='+', type=str, help='tasks to use in the construction of the vocabulary')
    parser.add_argument('--max_output_length', default=100, type=int, help='maximum output length for generation')
    parser.add_argument('--max_generative_vocab', default=50000, type=int,
                        help='max vocabulary for the generative softmax')
    parser.add_argument('--max_train_context_length', default=500, type=int,
                        help='maximum length of the contexts during training')
    parser.add_argument('--max_val_context_length', default=500, type=int,
                        help='maximum length of the contexts during validation')
    parser.add_argument('--max_answer_length', default=50, type=int,
                        help='maximum length of answers during training and validation')
    parser.add_argument('--subsample', default=20000000, type=int, help='subsample the datasets')
    parser.add_argument('--preserve_case', action='store_false', dest='lower',
                        help='whether to preserve casing for all text')
    parser.add_argument('--num_beams', type=int, default=1, help='number of beams to use for beam search')

    parser.add_argument('--model', type=str, choices=['Seq2Seq'], default='Seq2Seq', help='which model to import')
    parser.add_argument('--seq2seq_encoder', type=str, choices=['MQANEncoder', 'BiLSTM', 'Identity', 'Coattention'],
                        default='MQANEncoder', help='which encoder to use for the Seq2Seq model')
    parser.add_argument('--seq2seq_decoder', type=str, choices=['MQANDecoder'], default='MQANDecoder',
                        help='which decoder to use for the Seq2Seq model')
    parser.add_argument('--dimension', default=200, type=int, help='output dimensions for all layers')
    parser.add_argument('--rnn_dimension', default=None, type=int, help='output dimensions for RNN layers')
    parser.add_argument('--rnn_layers', default=1, type=int, help='number of layers for RNN modules')
    parser.add_argument('--rnn_zero_state', default='zero', choices=['zero', 'average'],
                        help='how to construct RNN zero state (for Identity encoder)')
    parser.add_argument('--transformer_layers', default=2, type=int, help='number of layers for transformer modules')
    parser.add_argument('--transformer_hidden', default=150, type=int, help='hidden size of the transformer modules')
    parser.add_argument('--transformer_heads', default=3, type=int, help='number of heads for transformer modules')
    parser.add_argument('--dropout_ratio', default=0.2, type=float, help='dropout for the model')

    parser.add_argument('--encoder_embeddings', default='glove+char',
                        help='which word embedding to use on the encoder side; use a bert-* pretrained model for BERT; or a xlm-roberta* model for Multi-lingual RoBERTa; '
                             'multiple embeddings can be concatenated with +; use @0, @1 to specify untied copies')
    parser.add_argument('--context_embeddings', default=None,
                        help='which word embedding to use for the context; use a bert-* pretrained model for BERT; '
                             'multiple embeddings can be concatenated with +; use @0, @1 to specify untied copies')
    parser.add_argument('--question_embeddings', default=None,
                        help='which word embedding to use for the question; use a bert-* pretrained model for BERT; '
                             'multiple embeddings can be concatenated with +; use @0, @1 to specify untied copies')
    parser.add_argument('--train_encoder_embeddings', action='store_true', default=False,
                        help='back propagate into pretrained encoder embedding (recommended for BERT and XLM-RoBERTa)')
    parser.add_argument('--train_context_embeddings', action='store_true', default=None,
                        help='back propagate into pretrained context embedding (recommended for BERT and XLM-RoBERTa)')
    parser.add_argument('--train_context_embeddings_after', type=int, default=0,
                        help='back propagate into pretrained context embedding after the given iteration (default: '
                             'immediately)')
    parser.add_argument('--train_question_embeddings', action='store_true', default=None,
                        help='back propagate into pretrained question embedding (recommended for BERT)')
    parser.add_argument('--train_question_embeddings_after', type=int, default=0,
                        help='back propagate into pretrained context embedding after the given iteration (default: '
                             'immediately)')
    parser.add_argument('--decoder_embeddings', default='glove+char',
                        help='which pretrained word embedding to use on the decoder side')
    parser.add_argument('--trainable_encoder_embeddings', default=0, type=int,
                        help='size of trainable portion of encoder embedding (only for Coattention encoder)')
    parser.add_argument('--trainable_decoder_embeddings', default=0, type=int,
                        help='size of trainable portion of decoder embedding (0 or omit to disable)')
    parser.add_argument('--pretrain_context', default=0, type=int,
                        help='number of pretraining steps for the context encoder')
    parser.add_argument('--pretrain_mlm_probability', default=0.15, type=int,
                        help='probability of replacing a token with mask for MLM pretraining')
    parser.add_argument('--force_subword_tokenize', action='store_true', default=False,
                        help='force subword tokenization of code tokens too')

    parser.add_argument('--warmup', default=800, type=int, help='warmup for learning rate')
    parser.add_argument('--grad_clip', default=1.0, type=float, help='gradient clipping')
    parser.add_argument('--beta0', default=0.9, type=float,
                        help='alternative momentum for Adam (only when not using transformer_lr)')
    parser.add_argument('--optimizer', default='adam', choices=['adam', 'sgd', 'radam'], type=str,
                        help='optimizer to use')
    parser.add_argument('--no_transformer_lr', action='store_false', dest='transformer_lr',
                        help='turns off the transformer learning rate strategy')
    parser.add_argument('--transformer_lr_multiply', default=1.0, type=float,
                        help='multiplier for transformer learning rate (if using Adam)')
    parser.add_argument('--lr_rate', default=0.001, type=float, help='fixed learning rate (if not using warmup)')
    parser.add_argument('--weight_decay', default=0.0, type=float, help='weight L2 regularization')
    parser.add_argument('--gradient_accumulation_steps', default=1, type=int, help='Number of accumulation steps. Useful to effectively get larger batch sizes.')
    

    parser.add_argument('--load', default=None, type=str, help='path to checkpoint to load model from inside args.save')
    parser.add_argument('--resume', action='store_true', help='whether to resume training with past optimizers')

    parser.add_argument('--seed', default=123, type=int, help='Random seed.')
    parser.add_argument('--devices', default=[0], nargs='+', type=int,
                        help='a list of devices that can be used for training')

    parser.add_argument('--no_commit', action='store_false', dest='commit',
                        help='do not track the git commit associated with this training run')
    parser.add_argument('--exist_ok', action='store_true',
                        help='Ok if the save directory already exists, i.e. overwrite is ok')

    parser.add_argument('--skip_cache', action='store_true',
                        help='whether to use exisiting cached splits or generate new ones')
    parser.add_argument('--use_curriculum', action='store_true', help='Use curriculum learning')
    parser.add_argument('--aux_dataset', default='', type=str,
                        help='path to auxiliary dataset (ignored if curriculum is not used)')
    parser.add_argument('--curriculum_max_frac', default=1.0, type=float,
                        help='max fraction of harder dataset to keep for curriculum')
    parser.add_argument('--curriculum_rate', default=0.1, type=float, help='growth rate for curriculum')
    parser.add_argument('--curriculum_strategy', default='linear', type=str, choices=['linear', 'exp'],
                        help='growth strategy for curriculum')


def post_parse(args):
    if args.val_task_names is None:
        args.val_task_names = []
        for t in args.train_task_names:
            if t not in args.val_task_names:
                args.val_task_names.append(t)
    if 'imdb' in args.val_task_names:
        args.val_task_names.remove('imdb')

    args.timestamp = datetime.datetime.now(tz=datetime.timezone.utc).isoformat()

    if len(args.train_task_names) > 1:
        if args.train_iterations is None:
            args.train_iterations = [1]
        if len(args.train_iterations) < len(args.train_task_names):
            args.train_iterations = len(args.train_task_names) * args.train_iterations
        if len(args.train_batch_tokens) < len(args.train_task_names):
            args.train_batch_tokens = len(args.train_task_names) * args.train_batch_tokens
    if len(args.val_batch_size) < len(args.val_task_names):
        args.val_batch_size = len(args.val_task_names) * args.val_batch_size

    # postprocess arguments
    if args.commit:
        args.commit = get_commit()
    else:
        args.commit = ''

    if args.rnn_dimension is None:
        args.rnn_dimension = args.dimension

    if args.context_embeddings is None:
        args.context_embeddings = args.encoder_embeddings
    if args.question_embeddings is None:
        args.question_embeddings = args.context_embeddings
    if args.train_context_embeddings is None:
        args.train_context_embeddings = args.train_encoder_embeddings
    if args.train_question_embeddings is None:
        args.train_question_embeddings = args.train_encoder_embeddings

    args.log_dir = args.save
    if args.tensorboard_dir is None:
        args.tensorboard_dir = args.log_dir
    args.dist_sync_file = os.path.join(args.log_dir, 'distributed_sync_file')

    for x in ['data', 'save', 'embeddings', 'log_dir', 'dist_sync_file']:
        setattr(args, x, os.path.join(args.root, getattr(args, x)))
    save_args(args)

    # create the task objects after we saved the configuration to the JSON file, because
    # tasks are not JSON serializable
    args.train_tasks = get_tasks(args.train_task_names, args)
    args.val_tasks = get_tasks(args.val_task_names, args)

    return args



#
# Copyright (c) 2019-2020 The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# This file was partially copied from huggingface's tokenizers library
# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from transformers import BertTokenizer, XLMRobertaTokenizer
from collections import OrderedDict


class MaskedXLMRobertaWordPieceTokenizer(object):
    def __init__(self, vocab, spm, added_tokens_encoder, added_tokens_decoder, unk_token, max_input_chars_per_word=100):
        self.vocab = vocab
        self.spm = spm
        self.unk_token = unk_token
        self.max_input_chars_per_word = max_input_chars_per_word
        self.added_tokens_encoder = added_tokens_encoder
        self.added_tokens_decoder = added_tokens_decoder

    def tokenize(self, tokens, mask):
        output_tokens = []
        for token, should_word_split in zip(tokens, mask):
            if not should_word_split:
                if token not in self.vocab and token not in self.added_tokens_encoder:
                    token_id = len(self.vocab) + len(self.added_tokens_encoder)
                    self.added_tokens_encoder[token] = token_id
                    self.added_tokens_decoder[token_id] = token
                output_tokens.append(token)
                continue

            chars = list(token)
            if len(chars) > self.max_input_chars_per_word:
                output_tokens.append(self.unk_token)
                continue

            sub_tokens = self.spm.EncodeAsPieces(token)
            output_tokens.extend(sub_tokens)

        return output_tokens


class MaskedBertWordPieceTokenizer(object):
    def __init__(self, vocab, added_tokens_encoder, added_tokens_decoder, unk_token, max_input_chars_per_word=100):
        self.vocab = vocab
        self.unk_token = unk_token
        self.max_input_chars_per_word = max_input_chars_per_word
        self.added_tokens_encoder = added_tokens_encoder
        self.added_tokens_decoder = added_tokens_decoder

    def tokenize(self, tokens, mask):
        output_tokens = []
        for token, should_word_split in zip(tokens, mask):
            if not should_word_split:
                if token not in self.vocab and token not in self.added_tokens_encoder:
                    token_id = len(self.vocab) + len(self.added_tokens_encoder)
                    self.added_tokens_encoder[token] = token_id
                    self.added_tokens_decoder[token_id] = token
                output_tokens.append(token)
                continue

            chars = list(token)
            if len(chars) > self.max_input_chars_per_word:
                output_tokens.append(self.unk_token)
                continue

            is_bad = False
            start = 0
            sub_tokens = []
            while start < len(chars):
                end = len(chars)
                cur_substr = None
                while start < end:
                    substr = "".join(chars[start:end])
                    if start > 0:
                        substr = "##" + substr
                    if substr in self.vocab:
                        cur_substr = substr
                        break
                    end -= 1
                if cur_substr is None:
                    is_bad = True
                    break
                sub_tokens.append(cur_substr)
                start = end

            if is_bad:
                output_tokens.append(self.unk_token)
            else:
                output_tokens.extend(sub_tokens)
        return output_tokens



class MaskedXLMRobertaTokenizer(XLMRobertaTokenizer):
    def __init__(self, *args, do_lower_case=False, do_basic_tokenize=False, **kwargs):
        # override do_lower_case and do_basic_tokenize unconditionally
        super().__init__(*args, do_lower_case=False, do_basic_tokenize=False, **kwargs)

        vocabs = ['<pad>'] + [self.sp_model.id_to_piece(i) for i in range(self.sp_model.get_piece_size())]
        self.vocab = OrderedDict((vocab, self._convert_token_to_id(vocab)) for vocab in vocabs)
        self.ids_to_tokens = OrderedDict((i, vocab) for vocab, i in self.vocab.items())

        # replace the word piece tokenizer with ours
        self.wordpiece_tokenizer = MaskedXLMRobertaWordPieceTokenizer(vocab=self.vocab,
                                                                      spm=self.sp_model,
                                                                      added_tokens_encoder=self.added_tokens_encoder,
                                                                      added_tokens_decoder=self.added_tokens_decoder,
                                                                      unk_token=self.unk_token)

        self._itos = IToSWrapper(self.ids_to_tokens, self.added_tokens_decoder)
        self._stoi = SToIWrapper(self.vocab, self.added_tokens_encoder)

    def tokenize(self, tokens, mask=None):
        return self.wordpiece_tokenizer.tokenize(tokens, mask)

    # provide an interface similar to Vocab

    def __len__(self):
        return len(self.vocab) + len(self.added_tokens_encoder)

    @property
    def stoi(self):
        return self._stoi

    @property
    def itos(self):
        return self._itos


class MaskedBertTokenizer(BertTokenizer):
    """
    A modified BertTokenizer that respects a mask deciding whether a token should be split or not.
    """

    def __init__(self, *args, do_lower_case=False, do_basic_tokenize=False, **kwargs):
        # override do_lower_case and do_basic_tokenize unconditionally
        super().__init__(*args, do_lower_case=False, do_basic_tokenize=False, **kwargs)

        # replace the word piece tokenizer with ours
        self.wordpiece_tokenizer = MaskedBertWordPieceTokenizer(vocab=self.vocab,
                                                            added_tokens_encoder=self.added_tokens_encoder,
                                                            added_tokens_decoder=self.added_tokens_decoder,
                                                            unk_token=self.unk_token)

        self._itos = IToSWrapper(self.ids_to_tokens, self.added_tokens_decoder)
        self._stoi = SToIWrapper(self.vocab, self.added_tokens_encoder)

    def tokenize(self, tokens, mask=None):
        return self.wordpiece_tokenizer.tokenize(tokens, mask)

    # provide an interface similar to Vocab

    def __len__(self):
        return len(self.vocab) + len(self.added_tokens_encoder)

    @property
    def stoi(self):
        return self._stoi

    @property
    def itos(self):
        return self._itos


class IToSWrapper:
    """Wrap the ordered dict vocabs to look like a list int -> str"""

    def __init__(self, base_vocab, added_tokens):
        self.base_vocab = base_vocab
        self.added_tokens = added_tokens

    def __getitem__(self, key):
        if isinstance(key, slice):
            return [self[key] for key in range(key.start or 0, key.stop or len(self), key.step or 1)]

        if key < len(self.base_vocab):
            return self.base_vocab[key]
        else:
            return self.added_tokens[key]

    def __len__(self):
        return len(self.base_vocab) + len(self.added_tokens)

    def __iter__(self):
        for key in range(len(self.base_vocab)):
            yield self.base_vocab[key]
        for key in range(len(self.base_vocab), len(self.base_vocab) + len(self.added_tokens)):
            yield self.added_tokens[key]


class SToIWrapper:
    """Wrap the ordered dict vocabs to look like a single dictionary"""

    def __init__(self, base_vocab, added_tokens):
        self.base_vocab = base_vocab
        self.added_tokens = added_tokens

    def __getitem__(self, key):
        if key in self.base_vocab:
            return self.base_vocab[key]
        else:
            return self.added_tokens[key]

    def __len__(self):
        return len(self.base_vocab) + len(self.added_tokens)

    def __iter__(self):
        for key in self.base_vocab:
            yield key
        for key in self.added_tokens:
            yield key



#
# Copyright (c) 2019, The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .base_task import BaseTask
from .registry import register_task
from . import generic_dataset


@register_task('multi30k')
class Multi30K(BaseTask):
    @property
    def metrics(self):
        return ['bleu', 'em', 'nem', 'nf1']

    def get_splits(self, root, **kwargs):
        src, trg = ['.' + x for x in self.name.split('.')[1:]]
        return generic_dataset.Multi30k.splits(exts=(src, trg),
                                               root=root,
                                               tokenize=self.tokenize,
                                               **kwargs)


@register_task('iwslt')
class IWSLT(BaseTask):
    @property
    def metrics(self):
        return ['bleu', 'em', 'nem', 'nf1']

    def get_splits(self, root, **kwargs):
        src, trg = ['.' + x for x in self.name.split('.')[1:]]
        return generic_dataset.IWSLT.splits(exts=(src, trg),
                                            root=root,
                                            tokenize=self.tokenize,
                                            **kwargs)


@register_task('squad')
class SQuAD(BaseTask):
    @property
    def metrics(self):
        return ['nf1', 'em', 'nem']

    def tokenize(self, sentence, field_name=None):
        if not sentence:
            return [], None
        return sentence.split(), None

    def detokenize(self, tokenized, field_name=None):
        return ' '.join(tokenized)

    def get_splits(self, root, **kwargs):
        return generic_dataset.SQuAD.splits(root=root,
                                            description=self.name,
                                            tokenize=self.tokenize,
                                            **kwargs)


@register_task('wikisql')
class WikiSQL(BaseTask):
    @property
    def metrics(self):
        return ['lfem', 'em', 'nem', 'nf1']

    def get_splits(self, root, **kwargs):
        return generic_dataset.WikiSQL.splits(
            root=root,
            query_as_question='query_as_question' in self.name,
            tokenize=self.tokenize,
            **kwargs)


@register_task('ontonotes')
class OntoNotesNER(BaseTask):
    def get_splits(self, root, **kwargs):
        split_task = self.name.split('.')
        _, _, subtask, nones, counting = split_task
        return generic_dataset.OntoNotesNER.splits(
            subtask=subtask, nones=True if nones == 'nones' else False,
            root=root,
            tokenize=self.tokenize,
            **kwargs)


@register_task('woz')
class WoZ(BaseTask):
    @property
    def metrics(self):
        return ['joint_goal_em', 'turn_request_em', 'turn_goal_em', 'avg_dialogue', 'em', 'nem', 'nf1']

    def get_splits(self, root, **kwargs):
        return generic_dataset.WOZ.splits(description=self.name,
                                          root=root,
                                          tokenize=self.tokenize,
                                          **kwargs)


@register_task('multinli')
class MultiNLI(BaseTask):
    def get_splits(self, root, **kwargs):
        return generic_dataset.MultiNLI.splits(description=self.name,
                                               root=root,
                                               tokenize=self.tokenize,
                                               **kwargs)


@register_task('srl')
class SRL(BaseTask):
    @property
    def metrics(self):
        return ['nf1', 'em', 'nem']

    def get_splits(self, root, **kwargs):
        return generic_dataset.SRL.splits(root=root, tokenize=self.tokenize, **kwargs)


@register_task('snli')
class SNLI(BaseTask):
    def get_splits(self, root, **kwargs):
        return generic_dataset.SNLI.splits(root=root, tokenize=self.tokenize, **kwargs)


@register_task('schema')
class WinogradSchema(BaseTask):
    def get_splits(self, root, **kwargs):
        return generic_dataset.WinogradSchema.splits(root=root, tokenize=self.tokenize, **kwargs)


class BaseSummarizationTask(BaseTask):
    @property
    def metrics(self):
        return ['avg_rouge', 'rouge1', 'rouge2', 'rougeL', 'em', 'nem', 'nf1']

    def preprocess_example(self, ex, train=False, max_context_length=None):
        # Filter examples with a dummy summary
        if train and 'This page includes the show' in ex.answer:
            return None

        return ex._replace(context=ex.context[:max_context_length])


@register_task('cnn')
class CNN(BaseSummarizationTask):
    def get_splits(self, root, **kwargs):
        return generic_dataset.CNN.splits(root=root, tokenize=self.tokenize, **kwargs)


@register_task('dailymail')
class DailyMail(BaseSummarizationTask):
    def get_splits(self, root, **kwargs):
        return generic_dataset.DailyMail.splits(root=root, tokenize=self.tokenize, **kwargs)


@register_task('cnn_dailymail')
class CNNDailyMail(BaseSummarizationTask):
    def get_splits(self, root, **kwargs):
        split_cnn = generic_dataset.CNN.splits(root=root, tokenize=self.tokenize, **kwargs)
        split_dm = generic_dataset.DailyMail.splits(root=root, tokenize=self.tokenize, **kwargs)
        for scnn, sdm in zip(split_cnn, split_dm):
            scnn.examples.extend(sdm)
        return split_cnn


@register_task('sst')
class SST(BaseTask):
    def get_splits(self, root, **kwargs):
        return generic_dataset.SST.splits(root=root, tokenize=self.tokenize, **kwargs)


@register_task('imdb')
class IMDB(BaseTask):
    def preprocess_example(self, ex, train=False, max_context_length=None):
        return ex._replace(context=ex.context[:max_context_length])

    def get_splits(self, root, **kwargs):
        kwargs['validation'] = None
        return generic_dataset.IMDb.splits(root=root, tokenize=self.tokenize, **kwargs)


@register_task('zre')
class ZRE(BaseTask):
    @property
    def metrics(self):
        return ['corpus_f1', 'precision', 'recall', 'em', 'nem', 'nf1']

    def get_splits(self, root, **kwargs):
        return generic_dataset.ZeroShotRE.splits(root=root, tokenize=self.tokenize, **kwargs)



#
# Copyright (c) 2018-2019, Salesforce, Inc.
#                          The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import torch

from .word_vectors import Vectors

ENTITIES = ['DATE', 'DURATION', 'EMAIL_ADDRESS', 'HASHTAG',
            'LOCATION', 'NUMBER', 'PHONE_NUMBER', 'QUOTED_STRING',
            'TIME', 'URL', 'USERNAME', 'PATH_NAME', 'CURRENCY']
MAX_ARG_VALUES = 5


class AlmondEmbeddings(Vectors):
    def __init__(self, name=None, cache=None, **kw):
        super().__init__(name, cache, **kw)

    def cache(self, name, cache, url=None):
        del name
        del cache

        dim = len(ENTITIES) + MAX_ARG_VALUES

        itos = []
        vectors = []
        for i, entity in enumerate(ENTITIES):
            for j in range(MAX_ARG_VALUES):
                itos.append(entity + '_' + str(j))
                vec = torch.zeros((dim,), dtype=torch.float32)
                vec[i] = 1.0
                vec[len(ENTITIES) + j] = 1.0
                vectors.append(vec)

        self.itos = itos
        self.stoi = {word: i for i, word in enumerate(itos)}
        self.vectors = torch.stack(vectors, dim=0).view(-1, dim)
        self.dim = dim



#
# Copyright (c) 2018, Salesforce, Inc.
#                     The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import json
import logging
import os
import random
import time

import numpy as np
import torch
import torch.nn.functional as F

from .data_utils.example import Batch
from .data_utils.iterator import Iterator

logger = logging.getLogger(__name__)


def tokenizer(s):
    return s.split()

def detokenize(text):
    tokens = ["'d", "n't", "'ve", "'m", "'re", "'ll", ".", ",", "?", "'s", ")"]
    for t in tokens:
        text = text.replace(' ' + t, t)
    text = text.replace("( ", "(")
    text = text.replace('gon na', 'gonna')
    text = text.replace('wan na', 'wanna')
    return text

def get_number_of_lines(file_path):
    count = 0
    with open(file_path) as f:
        for line in f:
            count += 1
    return count

def get_file_part_path(file_path, part_idx):
    return file_path + '_part' + str(part_idx+1)

def split_file_on_disk(file_path, num_splits):
    """
    """
    number_of_lines = get_number_of_lines(file_path)

    all_output_paths = []
    all_output_files = []
    for i in range(num_splits):
        output_path = get_file_part_path(file_path, i)
        all_output_paths.append(output_path)
        all_output_files.append(open(output_path, 'w'))

    written_lines = 0
    with open(file_path, 'r') as input_file:
        output_file_idx = 0
        for line in input_file:
            all_output_files[output_file_idx].write(line)
            written_lines += 1
            if written_lines % (number_of_lines//num_splits) == 0:
                output_file_idx = min(output_file_idx + 1, len(all_output_files)-1)

    for f in all_output_files:
        f.close()

    return all_output_paths

def combine_files_on_disk(file_path_prefix, num_files):
    with open(file_path_prefix, 'w') as combined_file:
        for i in range(num_files):
            file_path = get_file_part_path(file_path_prefix, i)
            with open(file_path, 'r') as file:
                for line in file:
                    combined_file.write(line)

def top_k_top_p_filtering(logits, top_k=0, top_p=1.0, filter_value=-float('Inf'), min_tokens_to_keep=1):
    """ Filter a distribution of logits using top-k and/or nucleus (top-p) filtering
        Args:
            logits: logits distribution shape (batch size, vocabulary size)
            if top_k > 0: keep only top k tokens with highest probability (top-k filtering).
            if top_k < 0: keeps all tokens but the ones with top |k| probability
            if top_p < 1.0: keep the top tokens with cumulative probability >= top_p (nucleus filtering).
                Nucleus filtering is described in Holtzman et al. (http://arxiv.org/abs/1904.09751)
            Make sure we keep at least min_tokens_to_keep per batch example in the output
        From: https://gist.github.com/thomwolf/1a5a29f6962089e871b94cbd09daf317
    """
    sign = 1
    if top_k < 0:
        top_k = logits.size(-1) + top_k
        sign = -1
    top_k = min(top_k, logits.size(-1))  # Safety check
    if top_k > 0:
        # Remove all tokens with a probability less than the last token of the top-k
        indices_to_remove = sign*logits < torch.topk(sign*logits, top_k)[0][..., -1, None]
        logits[indices_to_remove] = filter_value

    if top_p < 1.0:
        sorted_logits, sorted_indices = torch.sort(logits, descending=True)
        cumulative_probs = torch.cumsum(F.softmax(sorted_logits, dim=-1), dim=-1)

        # Remove tokens with cumulative probability above the threshold
        sorted_indices_to_remove = cumulative_probs > top_p
        if min_tokens_to_keep > 1:
            # Keep at least min_tokens_to_keep (set to min_tokens_to_keep-1 because we add the first one below)
            sorted_indices_to_remove[..., :min_tokens_to_keep] = 0
        # Shift the indices to the right to keep also the first token above the threshold
        sorted_indices_to_remove[..., 1:] = sorted_indices_to_remove[..., :-1].clone()
        sorted_indices_to_remove[..., 0] = 0

        # scatter sorted tensors to original indexing
        indices_to_remove = sorted_indices_to_remove.scatter(dim=1, index=sorted_indices, src=sorted_indices_to_remove)
        logits[indices_to_remove] = filter_value
    return logits

def map_filter(callable, iterable):
    output = []
    for element in iterable:
        new_element = callable(element)
        if new_element is not None:
            output.append(new_element)
    return output


def preprocess_examples(args, tasks, splits, logger=None, train=True):
    min_length = 1
    max_context_length = args.max_train_context_length if train else args.max_val_context_length
    is_too_long = lambda ex: (len(ex.answer) > args.max_answer_length or
                              len(ex.context) > max_context_length)
    is_too_short = lambda ex: (len(ex.answer) < min_length or
                               len(ex.context) < min_length)

    for task, s in zip(tasks, splits):
        if logger is not None:
            logger.info(f'{task.name} has {len(s.examples)} examples')

        l = len(s.examples)
        s.examples = map_filter(
            lambda ex: task.preprocess_example(ex, train=train, max_context_length=max_context_length),
            s.examples)

        if train:
            l = len(s.examples)
            s.examples = [ex for ex in s.examples if not is_too_long(ex)]
            if len(s.examples) < l:
                if logger is not None:
                    logger.info(f'Filtering out long {task.name} examples: {l} -> {len(s.examples)}')

            l = len(s.examples)
            s.examples = [ex for ex in s.examples if not is_too_short(ex)]
            if len(s.examples) < l:
                if logger is not None:
                    logger.info(f'Filtering out short {task.name} examples: {l} -> {len(s.examples)}')

        if logger is not None:
            context_lengths = [len(ex.context) for ex in s.examples]
            question_lengths = [len(ex.question) for ex in s.examples]
            answer_lengths = [len(ex.answer) for ex in s.examples]

            logger.info(
                f'{task.name} context lengths (min, mean, max): {np.min(context_lengths)}, {int(np.mean(context_lengths))}, {np.max(context_lengths)}')
            logger.info(
                f'{task.name} question lengths (min, mean, max): {np.min(question_lengths)}, {int(np.mean(question_lengths))}, {np.max(question_lengths)}')
            logger.info(
                f'{task.name} answer lengths (min, mean, max): {np.min(answer_lengths)}, {int(np.mean(answer_lengths))}, {np.max(answer_lengths)}')

        if logger is not None:
            logger.info('Tokenized examples:')
            for ex in s.examples[:10]:
                logger.info('Context: ' + ' '.join([token.strip() for token in ex.context]))
                logger.info('Question: ' + ' '.join([token.strip() for token in ex.question]))
                logger.info('Answer: ' + ' '.join([token.strip() for token in ex.answer]))


def init_devices(args, devices=None):
    if not torch.cuda.is_available():
        return [torch.device('cpu')]
    if not devices:
        return [torch.device('cuda:0')]
    return [torch.device(ordinal) for ordinal in devices]


def set_seed(args):
    np.random.seed(args.seed)
    random.seed(args.seed)
    torch.manual_seed(args.seed)
    torch.cuda.manual_seed_all(args.seed)


def get_trainable_params(model, name=False):
    if name:
        return list(filter(lambda p: p[1].requires_grad, model.named_parameters()))
    else:
        return list(filter(lambda p: p.requires_grad, model.parameters()))


def log_model_size(logger, model, model_name):
    num_param = sum([p.nelement() for p in model.parameters() if p.requires_grad])
    logger.info(f'{model_name} has {num_param:,} parameters')


def elapsed_time(log):
    t = time.time() - log.start
    day = int(t // (24 * 3600))
    t = t % (24 * 3600)
    hour = int(t // 3600)
    t %= 3600
    minutes = int(t // 60)
    t %= 60
    seconds = int(t)
    return f'{day:02}:{hour:02}:{minutes:02}:{seconds:02}'


def batch_fn(new, i, sofar):
    prev_max_len = sofar / (i - 1) if i > 1 else 0
    return max(len(new.context), 5 * len(new.answer), prev_max_len) * i


def make_data_loader(dataset, numericalizer, batch_size, device=None, train=False):
    iterator = Iterator(dataset, batch_size,
                        batch_size_fn=batch_fn if train else None,
                        shuffle=train,
                        repeat=train,
                        bucket_by_sort_key=train)
    return torch.utils.data.DataLoader(iterator, batch_size=None,
                                       collate_fn=lambda minibatch: Batch.from_examples(minibatch, numericalizer,
                                                                                        device=device))


def pad(x, new_channel, dim, val=None):
    if x.size(dim) > new_channel:
        x = x.narrow(dim, 0, new_channel)
    channels = x.size()
    assert (new_channel >= channels[dim])
    if new_channel == channels[dim]:
        return x
    size = list(channels)
    size[dim] = new_channel - size[dim]
    padding = x.new(*size).fill_(val)
    return torch.cat([x, padding], dim)


def load_config_json(args):
    args.almond_type_embeddings = False
    with open(os.path.join(args.path, 'config.json')) as config_file:
        config = json.load(config_file)
        retrieve = ['model', 'seq2seq_encoder', 'seq2seq_decoder', 'transformer_layers', 'rnn_layers', 'rnn_zero_state',
                    'transformer_hidden', 'dimension', 'rnn_dimension', 'load', 'max_val_context_length',
                    'val_batch_size', 'transformer_heads', 'max_output_length', 'max_generative_vocab', 'lower',
                    'encoder_embeddings', 'context_embeddings', 'question_embeddings', 'decoder_embeddings',
                    'trainable_decoder_embeddings', 'trainable_encoder_embeddings', 'train_encoder_embeddings',
                    'train_context_embeddings', 'train_question_embeddings', 'locale', 'use_pretrained_bert',
                    'train_context_embeddings_after', 'train_question_embeddings_after',
                    'pretrain_context', 'pretrain_mlm_probability', 'force_subword_tokenize', 'num_beams']

        for r in retrieve:
            if r in config:
                setattr(args, r, config[r])
            # These are for backward compatibility with models that were trained before we added these arguments
            elif r == 'locale':
                setattr(args, r, 'en')
            elif r in ('trainable_decoder_embedding', 'trainable_encoder_embeddings', 'pretrain_context',
                       'train_context_embeddings_after', 'train_question_embeddings_after'):
                setattr(args, r, 0)
            elif r == 'pretrain_mlm_probability':
                setattr(args, r, 0.15)
            elif r == 'context_embeddings':
                if args.seq2seq_encoder == 'Coattention':
                    setattr(args, r, '')
                else:
                    setattr(args, r, args.encoder_embeddings)
            elif r == 'question_embeddings':
                setattr(args, r, args.encoder_embeddings)
            elif r == 'train_encoder_embeddings':
                setattr(args, r, False)
            elif r == 'train_context_embeddings':
                if args.seq2seq_encoder == 'Coattention':
                    setattr(args, r, False)
                else:
                    setattr(args, r, args.train_encoder_embeddings)
            elif r == 'train_question_embeddings':
                setattr(args, r, args.train_encoder_embeddings)
            elif r == 'rnn_dimension':
                setattr(args, r, args.dimension)
            elif r == 'rnn_zero_state':
                setattr(args, r, 'zero')
            elif r == 'use_pretrained_bert':
                setattr(args, r, True)
            elif r == 'num_beams':
                setattr(args, r, 1)
            else:
                setattr(args, r, None)
        args.dropout_ratio = 0.0

    args.best_checkpoint = os.path.join(args.path, args.checkpoint_name)



#
# Copyright (c) 2018, Salesforce, Inc.
#                     The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import torch

from .coatt_encoder import CoattentionEncoder
from .lstm_encoder import BiLSTMEncoder
from .mqan_encoder import MQANEncoder
from .identity_encoder import IdentityEncoder
from .mqan_decoder import MQANDecoder
from .common import mask_tokens

ENCODERS = {
    'MQANEncoder': MQANEncoder,
    'BiLSTM': BiLSTMEncoder,
    'Identity': IdentityEncoder,
    'Coattention': CoattentionEncoder,
}
DECODERS = {
    'MQANDecoder': MQANDecoder
}


class Seq2Seq(torch.nn.Module):
    def __init__(self, numericalizer, args, context_embeddings, question_embeddings, decoder_embeddings):
        super().__init__()
        self.args = args

        self.numericalizer = numericalizer
        self.encoder = ENCODERS[args.seq2seq_encoder](numericalizer, args, context_embeddings, question_embeddings)
        self.decoder = DECODERS[args.seq2seq_decoder](numericalizer, args, decoder_embeddings)

        if self.args.pretrain_context > 0:
            self.context_pretrain_lm_head = torch.nn.Linear(self.args.dimension, numericalizer.num_tokens)

    def set_train_context_embeddings(self, trainable):
        self.encoder.set_train_context_embeddings(trainable)

    def set_train_question_embeddings(self, trainable):
        self.encoder.set_train_question_embeddings(trainable)

    def _pretrain_forward(self, batch):
        masked_input, masked_labels = mask_tokens(batch.context.value, self.numericalizer,
                                                  self.args.pretrain_mlm_probability)
        masked_batch = batch._replace(context=batch.context._replace(value=masked_input))

        self_attended_context, _final_context, _context_rnn_state, _final_question, _question_rnn_state = \
            self.encoder(masked_batch)
        context_logits = self.context_pretrain_lm_head(self_attended_context[-1])
        predictions = None

        context_logits = context_logits.view(-1, self.numericalizer.num_tokens)
        masked_labels = masked_labels.view(-1)
        loss = torch.nn.functional.cross_entropy(context_logits, masked_labels, ignore_index=self.numericalizer.pad_id)
        return loss, predictions

    def _normal_forward(self, batch):
        self_attended_context, final_context, context_rnn_state, final_question, question_rnn_state = \
            self.encoder(batch)
        return self.decoder(batch, self_attended_context, final_context, context_rnn_state,
                            final_question, question_rnn_state)

    def forward(self, batch, iteration, pretraining=False):
        if pretraining:
            return self._pretrain_forward(batch)
        else:
            return self._normal_forward(batch)



#
# Copyright (c) 2018, Salesforce, Inc.
#                     The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import math
from typing import NamedTuple, List

import torch
import torch.nn as nn
from torch.autograd import Variable
from torch.jit import Final
from torch.nn import functional as F
from torch.nn.utils.rnn import pack_padded_sequence as pack
from torch.nn.utils.rnn import pad_packed_sequence as unpack


class EmbeddingOutput(NamedTuple):
    all_layers: List[torch.Tensor]
    last_layer: torch.Tensor


INF = 1e10
EPSILON = 1e-10


class MultiLSTMCell(nn.Module):
    def __init__(self, num_layers, input_size, rnn_size, dropout):
        super(MultiLSTMCell, self).__init__()
        self.dropout = nn.Dropout(dropout)
        self.num_layers = num_layers
        self.layers = nn.ModuleList()

        for i in range(num_layers):
            self.layers.append(nn.LSTMCell(input_size, rnn_size))
            input_size = rnn_size

    def forward(self, input, hidden):
        h_0, c_0 = hidden
        h_1, c_1 = [], []
        for i, layer in enumerate(self.layers):
            input = self.dropout(input)
            h_1_i, c_1_i = layer(input, (h_0[i], c_0[i]))
            input = h_1_i
            h_1 += [h_1_i]
            c_1 += [c_1_i]

        h_1 = torch.stack(h_1)
        c_1 = torch.stack(c_1)

        return input, (h_1, c_1)


def positional_encodings_like(x, t=None):
    if t is None:
        positions = torch.arange(0., x.size(1))
        if x.is_cuda:
            positions = positions.cuda(x.get_device())
    else:
        positions = t
    encodings = torch.zeros(*x.size()[1:])
    if x.is_cuda:
        encodings = encodings.cuda(x.get_device())
    for channel in range(x.size(-1)):
        if channel % 2 == 0:
            encodings[:, channel] = torch.sin(
                positions / 10000 ** (channel / x.size(2)))
        else:
            encodings[:, channel] = torch.cos(
                positions / 10000 ** ((channel - 1) / x.size(2)))
    return Variable(encodings)


# torch.matmul can't do (4, 3, 2) @ (4, 2) -> (4, 3)
def matmul(x, y):
    if x.dim() == y.dim():
        return x @ y
    if x.dim() == y.dim() - 1:
        return (x.unsqueeze(-2) @ y).squeeze(-2)
    return (x @ y.unsqueeze(-2)).squeeze(-2)


def pad_to_match(x, y):
    x_len, y_len = x.size(1), y.size(1)
    if x_len == y_len:
        return x, y
    extra = x.new_ones((x.size(0), abs(y_len - x_len)))
    if x_len < y_len:
        return torch.cat((x, extra), 1), y
    return x, torch.cat((y, extra), 1)


class LayerNorm(nn.Module):

    def __init__(self, d_model, eps=1e-6):
        super().__init__()
        self.gamma = nn.Parameter(torch.ones(d_model))
        self.beta = nn.Parameter(torch.zeros(d_model))
        self.eps = eps

    def forward(self, x):
        mean = x.mean(-1, keepdim=True)
        std = x.std(-1, keepdim=True)
        return self.gamma * (x - mean) / (std + self.eps) + self.beta


class ResidualBlock(nn.Module):

    def __init__(self, layer, d_model, dropout_ratio):
        super().__init__()
        self.layer = layer
        self.dropout = nn.Dropout(dropout_ratio)
        self.layernorm = LayerNorm(d_model)

    def forward(self, *x, padding=None):
        return self.layernorm(x[0] + self.dropout(self.layer(*x, padding=padding)))


class Attention(nn.Module):

    def __init__(self, d_key, dropout_ratio, causal):
        super().__init__()
        self.scale = math.sqrt(d_key)
        self.dropout = nn.Dropout(dropout_ratio)
        self.causal = causal

    def forward(self, query, key, value, padding=None):
        dot_products = matmul(query, key.transpose(1, 2))
        if query.dim() == 3 and self.causal:
            tri = key.new_ones((key.size(1), key.size(1))).triu(1) * INF
            dot_products.sub_(tri.unsqueeze(0))
        if padding is not None:
            if dot_products.dim() == 3:
                # dot_products is batch x query time x key time
                # padding is batch x key time
                # unsqueeze to batch x 1 x key time then broadcast
                dot_products.masked_fill_(padding.unsqueeze(1).expand_as(dot_products), -INF)
            else:
                # dot_products is batch x key time and is directly compatible with padding
                dot_products.masked_fill_(padding, -INF)
        return matmul(self.dropout(F.softmax(dot_products / self.scale, dim=-1)), value)


class MultiHead(nn.Module):

    def __init__(self, d_key, d_value, n_heads, dropout_ratio, causal=False):
        super().__init__()
        self.attention = Attention(d_key, dropout_ratio, causal=causal)
        self.wq = Linear(d_key, d_key, bias=False)
        self.wk = Linear(d_key, d_key, bias=False)
        self.wv = Linear(d_value, d_value, bias=False)
        self.n_heads = n_heads

    def forward(self, query, key, value, padding=None):
        query, key, value = self.wq(query), self.wk(key), self.wv(value)
        query, key, value = (
            x.chunk(self.n_heads, -1) for x in (query, key, value))
        return torch.cat([self.attention(q, k, v, padding=padding)
                          for q, k, v in zip(query, key, value)], -1)


class LinearReLU(nn.Module):

    def __init__(self, d_model, d_hidden):
        super().__init__()
        self.feedforward = Feedforward(d_model, d_hidden, activation='relu')
        self.linear = Linear(d_hidden, d_model)

    def forward(self, x, padding=None):
        return self.linear(self.feedforward(x))


class TransformerEncoderLayer(nn.Module):

    def __init__(self, dimension, n_heads, hidden, dropout):
        super().__init__()
        self.selfattn = ResidualBlock(
            MultiHead(
                dimension, dimension, n_heads, dropout),
            dimension, dropout)
        self.feedforward = ResidualBlock(
            LinearReLU(dimension, hidden),
            dimension, dropout)

    def forward(self, x, padding=None):
        return self.feedforward(self.selfattn(x, x, x, padding=padding))


class TransformerEncoder(nn.Module):

    def __init__(self, dimension, n_heads, hidden, num_layers, dropout):
        super().__init__()
        self.layers = nn.ModuleList(
            [TransformerEncoderLayer(dimension, n_heads, hidden, dropout) for i in range(num_layers)])
        self.dropout = nn.Dropout(dropout)

    def forward(self, x, padding=None):
        x = self.dropout(x)
        encoding = [x]
        for layer in self.layers:
            x = layer(x, padding=padding)
            encoding.append(x)
        return encoding


class TransformerDecoderLayer(nn.Module):

    def __init__(self, dimension, n_heads, hidden, dropout, causal=True):
        super().__init__()
        self.selfattn = ResidualBlock(
            MultiHead(dimension, dimension, n_heads,
                      dropout, causal),
            dimension, dropout)
        self.attention = ResidualBlock(
            MultiHead(dimension, dimension, n_heads,
                      dropout),
            dimension, dropout)
        self.feedforward = ResidualBlock(
            LinearReLU(dimension, hidden),
            dimension, dropout)

    def forward(self, x, encoding, selfattn_keys=None, context_padding=None, answer_padding=None):
        if selfattn_keys is None:
            selfattn_keys = x
        x = self.selfattn(x, selfattn_keys, selfattn_keys, padding=answer_padding)
        return self.feedforward(self.attention(x, encoding, encoding, padding=context_padding))


class TransformerDecoder(nn.Module):

    def __init__(self, dimension, n_heads, hidden, num_layers, dropout, causal=True):
        super().__init__()
        self.layers = nn.ModuleList(
            [TransformerDecoderLayer(dimension, n_heads, hidden, dropout, causal=causal) for i in range(num_layers)])
        self.dropout = nn.Dropout(dropout)
        self.d_model = dimension

    def forward(self, x, encoding, context_padding=None, positional_encodings=True, answer_padding=None):
        if positional_encodings:
            x = x + positional_encodings_like(x)
        x = self.dropout(x)
        for layer, enc in zip(self.layers, encoding[1:]):
            x = layer(x, enc, context_padding=context_padding, answer_padding=answer_padding)
        return x


def mask(targets, out, squash=True, pad_idx=1):
    mask = (targets != pad_idx)
    out_mask = mask.unsqueeze(-1).expand_as(out).contiguous()
    if squash:
        out_after = out[out_mask].contiguous().view(-1, out.size(-1))
    else:
        out_after = out * out_mask.float()
    targets_after = targets[mask]
    return out_after, targets_after


class Highway(torch.nn.Module):
    def __init__(self, d_in, activation='relu', n_layers=1):
        super(Highway, self).__init__()
        self.d_in = d_in
        self._layers = torch.nn.ModuleList([Linear(d_in, 2 * d_in) for _ in range(n_layers)])
        for layer in self._layers:
            layer.bias[d_in:].fill_(1)
        self.activation = getattr(F, activation)

    def forward(self, inputs):
        current_input = inputs
        for layer in self._layers:
            projected_input = layer(current_input)
            linear_part = current_input
            nonlinear_part = projected_input[:, :self.d_in] if projected_input.dim() == 2 \
                else projected_input[:, :, :self.d_in]
            nonlinear_part = self.activation(nonlinear_part)
            gate = projected_input[:, self.d_in:(2 * self.d_in)] if projected_input.dim() == 2 \
                else projected_input[:, :, self.d_in:(2 * self.d_in)]
            gate = F.sigmoid(gate)
            current_input = gate * linear_part + (1 - gate) * nonlinear_part
        return current_input


class LinearFeedforward(nn.Module):

    def __init__(self, d_in, d_hid, d_out, activation='relu', dropout=0.2):
        super().__init__()
        self.feedforward = Feedforward(d_in, d_hid, activation=activation)
        self.linear = Linear(d_hid, d_out)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        return self.dropout(self.linear(self.feedforward(x)))


class PackedLSTM(nn.Module):

    def __init__(self, d_in, d_out, bidirectional=False, num_layers=1,
                 dropout=0.0, batch_first=True):
        """A wrapper class that packs input sequences and unpacks output sequences"""
        super().__init__()
        if bidirectional:
            d_out = d_out // 2
        self.rnn = nn.LSTM(d_in, d_out,
                           num_layers=num_layers,
                           dropout=dropout,
                           bidirectional=bidirectional,
                           batch_first=batch_first)
        self.batch_first = batch_first

    def forward(self, inputs, lengths, hidden=None):
        lens, indices = torch.sort(lengths.clone().detach(), 0, True)
        inputs = inputs[indices] if self.batch_first else inputs[:, indices]
        outputs, (h, c) = self.rnn(pack(inputs, lens.tolist(),
                                        batch_first=self.batch_first), hidden)
        outputs = unpack(outputs, batch_first=self.batch_first)[0]
        _, _indices = torch.sort(indices, 0)
        outputs = outputs[_indices] if self.batch_first else outputs[:, _indices]
        h, c = h[:, _indices, :], c[:, _indices, :]
        return outputs, (h, c)


class Linear(nn.Linear):

    def forward(self, x):
        size = x.size()
        return super().forward(
            x.contiguous().view(-1, size[-1])).view(*size[:-1], -1)


class Feedforward(nn.Module):

    def __init__(self, d_in, d_out, activation=None, bias=True, dropout=0.2):
        super().__init__()
        if activation is not None:
            self.activation = getattr(torch, activation)
        else:
            self.activation = lambda x: x
        self.linear = Linear(d_in, d_out, bias=bias)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        return self.activation(self.linear(self.dropout(x)))


class CombinedEmbedding(nn.Module):
    project: Final[bool]
    dimension: Final[int]

    def __init__(self, numericalizer, pretrained_embeddings,
                 output_dimension,
                 finetune_pretrained=False,
                 trained_dimension=0,
                 project=True):
        super().__init__()
        self.project = project
        self.pretrained_embeddings = nn.ModuleList(pretrained_embeddings)

        dimension = 0
        for idx, embedding in enumerate(self.pretrained_embeddings):
            dimension += embedding.dim
        self.set_trainable(finetune_pretrained)

        if trained_dimension > 0:
            self.trained_embeddings = nn.Embedding(numericalizer.num_tokens, trained_dimension)
            dimension += trained_dimension
        else:
            self.trained_embeddings = None
        if self.project:
            self.projection = Feedforward(dimension, output_dimension)
        else:
            assert dimension == output_dimension, (dimension, output_dimension)
        self.dimension = output_dimension

    def set_trainable(self, trainable):
        self.pretrained_embeddings.requires_grad_(trainable)

    def _combine_embeddings(self, embeddings):
        if len(embeddings) == 1:
            all_layers = embeddings[0].all_layers
            last_layer = embeddings[0].last_layer
            if self.project:
                last_layer = self.projection(last_layer)
            return EmbeddingOutput(all_layers=all_layers, last_layer=last_layer)

        all_layers = None
        last_layer = []
        for emb in embeddings:
            if all_layers is None:
                all_layers = [[layer] for layer in emb.all_layers]
            elif len(all_layers) != len(emb.all_layers):
                raise ValueError('Cannot combine embeddings that use different numbers of layers')
            else:
                for layer_list, layer in zip(all_layers, emb.all_layers):
                    layer_list.append(layer)
            last_layer.append(emb.last_layer)

        all_layers = [torch.cat(layer, dim=2) for layer in all_layers]
        last_layer = torch.cat(last_layer, dim=2)
        if self.project:
            last_layer = self.projection(last_layer)
        return EmbeddingOutput(all_layers=all_layers, last_layer=last_layer)

    def forward(self, x, padding=None):
        embedded: List[EmbeddingOutput] = []
        if self.pretrained_embeddings is not None:
            embedded += [emb(x, padding=padding) for emb in self.pretrained_embeddings]

        if self.trained_embeddings is not None:
            trained_vocabulary_size = self.trained_embeddings.weight.size()[0]
            valid_x = torch.lt(x, trained_vocabulary_size)
            masked_x = torch.where(valid_x, x, torch.zeros_like(x))
            output = self.trained_embeddings(masked_x)
            embedded.append(EmbeddingOutput(all_layers=[output], last_layer=output))

        return self._combine_embeddings(embedded)


class SemanticFusionUnit(nn.Module):

    def __init__(self, d, l):
        super().__init__()
        self.r_hat = Feedforward(d * l, d, 'tanh')
        self.g = Feedforward(d * l, d, 'sigmoid')
        self.dropout = nn.Dropout(0.2)

    def forward(self, x):
        c = self.dropout(torch.cat(x, -1))
        r_hat = self.r_hat(c)
        g = self.g(c)
        o = g * r_hat + (1 - g) * x[0]
        return o


class LSTMDecoderAttention(nn.Module):
    def __init__(self, dim, dot=False):
        super().__init__()
        self.linear_in = nn.Linear(dim, dim, bias=False)
        self.linear_out = nn.Linear(2 * dim, dim, bias=False)
        self.tanh = nn.Tanh()
        self.mask = None
        self.dot = dot

    def applyMasks(self, context_mask):
        # context_mask is batch x encoder_time, convert it to batch x 1 x encoder_time
        self.context_mask = context_mask.unsqueeze(1)

    def forward(self, input: torch.Tensor, context: torch.Tensor):
        # input is batch x decoder_time x dim
        # context is batch x encoder_time x dim
        # output will be batch x decoder_time x dim
        # context_attention will be batch x decoder_time x encoder_time

        if not self.dot:
            targetT = self.linear_in(input)  # batch x decoder_time x dim x 1
        else:
            targetT = input

        transposed_context = torch.transpose(context, 2, 1)
        context_scores = torch.matmul(targetT, transposed_context)
        context_scores.masked_fill_(self.context_mask, -float('inf'))
        context_attention = F.softmax(context_scores, dim=-1) + EPSILON

        # convert context_attention to batch x decoder_time x 1 x encoder_time
        # convert context to batch x 1 x encoder_time x dim
        # context_alignment will be batch x decoder_time x 1 x dim
        context_alignment = torch.matmul(context_attention.unsqueeze(2), context.unsqueeze(1))
        # squeeze out the extra dimension
        context_alignment = context_alignment.squeeze(2)

        combined_representation = torch.cat([input, context_alignment], 2)
        output = self.tanh(self.linear_out(combined_representation))

        return output, context_attention


class CoattentiveLayer(nn.Module):

    def __init__(self, d, dropout=0.2):
        super().__init__()
        self.proj = Feedforward(d, d, dropout=0.0)
        self.embed_sentinel = nn.Embedding(2, d)
        self.dropout = nn.Dropout(dropout)

    def forward(self, context, question, context_padding, question_padding):
        context_padding = torch.cat([context.new_zeros((context.size(0), 1), dtype=torch.bool), context_padding], 1)
        question_padding = torch.cat([question.new_zeros((question.size(0), 1), dtype=torch.bool), question_padding], 1)

        context_sentinel = self.embed_sentinel(context.new_zeros((context.size(0), 1), dtype=torch.long))
        context = torch.cat([context_sentinel, self.dropout(context)], 1) # batch_size x (context_length + 1) x features

        question_sentinel = self.embed_sentinel(question.new_ones((question.size(0), 1), dtype=torch.long))
        question = torch.cat([question_sentinel, question], 1)  # batch_size x (question_length + 1) x features
        question = torch.tanh(self.proj(question))  # batch_size x (question_length + 1) x features

        affinity = context.bmm(question.transpose(1, 2))  # batch_size x (context_length + 1) x (question_length + 1)
        attn_over_context = self.normalize(affinity, context_padding)  # batch_size x (context_length + 1) x 1
        attn_over_question = self.normalize(affinity.transpose(1, 2),
                                            question_padding)  # batch_size x (question_length + 1) x 1
        sum_of_context = self.attn(attn_over_context, context)  # batch_size x (question_length + 1) x features
        sum_of_question = self.attn(attn_over_question, question)  # batch_size x (context_length + 1) x features
        coattn_context = self.attn(attn_over_question, sum_of_context)  # batch_size x (context_length + 1) x features
        coattn_question = self.attn(attn_over_context, sum_of_question)  # batch_size x (question_length + 1) x features
        return torch.cat([coattn_context, sum_of_question], 2)[:, 1:], \
               torch.cat([coattn_question, sum_of_context], 2)[:, 1:]

    @staticmethod
    def attn(weights, candidates):
        w1, w2, w3 = weights.size()
        c1, c2, c3 = candidates.size()
        return weights.unsqueeze(3).expand(w1, w2, w3, c3).mul(candidates.unsqueeze(2).expand(c1, c2, w3, c3)) \
            .sum(1).squeeze(1)

    @staticmethod
    def normalize(original, padding):
        raw_scores = original.clone()
        raw_scores.masked_fill_(padding.unsqueeze(-1).expand_as(raw_scores), -INF)
        return F.softmax(raw_scores, dim=1)


# The following was copied and adapted from Hugginface's Tokenizers library
# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

def mask_tokens(inputs: torch.Tensor, numericalizer, probability):
    """ Prepare masked tokens inputs/labels for masked language modeling: 80% MASK, 10% random, 10% original. """

    assert numericalizer.mask_id

    labels = inputs.clone()
    # We sample a few tokens in each sequence for masked-LM training
    probability_matrix = torch.full(labels.shape, probability, device=inputs.device)
    special_tokens_mask = [numericalizer.get_special_token_mask(token_ids) for token_ids in inputs.tolist()]
    probability_matrix.masked_fill_(torch.tensor(special_tokens_mask, dtype=torch.bool, device=inputs.device),
                                    value=0.0)
    padding_mask = labels.eq(numericalizer.pad_id)
    probability_matrix.masked_fill_(padding_mask, value=0.0)
    masked_indices = torch.bernoulli(probability_matrix).bool()

    # All tokens that are not masked become padding, so we don't compute loss on them
    labels[~masked_indices] = numericalizer.pad_id

    # 80% of the time, we replace masked input tokens with tokenizer.mask_token ([MASK])
    indices_replaced = torch.bernoulli(torch.full(labels.shape, 0.8, device=inputs.device)).bool() & masked_indices
    inputs[indices_replaced] = numericalizer.mask_id

    # 10% of the time, we replace masked input tokens with random word
    indices_random = torch.bernoulli(torch.full(labels.shape, 0.5, device=inputs.device)).bool() & masked_indices & ~indices_replaced
    random_words = torch.randint(numericalizer.num_tokens, labels.shape, dtype=torch.long, device=inputs.device)
    inputs[indices_random] = random_words[indices_random]

    # The rest of the time (10% of the time) we keep the masked input tokens unchanged
    return inputs, labels


class BeamHypotheses(object):

    def __init__(self, n_hyp, max_length, length_penalty, early_stopping):
        """
        Initialize n-best list of hypotheses.
        """
        self.max_length = max_length - 1  # ignoring bos_token
        self.length_penalty = length_penalty
        self.early_stopping = early_stopping
        self.n_hyp = n_hyp
        self.hyp = []
        self.worst_score = 1e9

    def __len__(self):
        """
        Number of hypotheses in the list.
        """
        return len(self.hyp)

    def add(self, hyp, sum_logprobs):
        """
        Add a new hypothesis to the list.
        """
        # print('hyp = ', hyp)
        # print('sum_logprobs = ', sum_logprobs)
        # print('len(hyp) ** self.length_penalty = ', len(hyp) ** self.length_penalty)
        score = sum_logprobs / len(hyp) ** self.length_penalty
        # print('score = ', score)
        if len(self) < self.n_hyp or score > self.worst_score:
            self.hyp.append((score, hyp))
            if len(self) > self.n_hyp:
                sorted_scores = sorted([(s, idx) for idx, (s, _) in enumerate(self.hyp)])
                del self.hyp[sorted_scores[0][1]]
                self.worst_score = sorted_scores[0][0]
            else:
                self.worst_score = min(score, self.worst_score)

    def is_done(self, best_sum_logprobs):
        """
        If there are enough hypotheses and that none of the hypotheses being generated
        can become better than the worst one in the heap, then we are done with this sentence.
        """
        if len(self) < self.n_hyp:
            return False
        elif self.early_stopping:
            return True
        else:
            return self.worst_score >= best_sum_logprobs / self.max_length ** self.length_penalty



import logging
from collections import defaultdict, Counter

logger = logging.getLogger(__name__)


def _default_unk_index():
    return 0


class Vocab(object):
    """Defines a vocabulary object that will be used to numericalize a field.

    Attributes:
        freqs: A collections.Counter object holding the frequencies of tokens
            in the data used to build the Vocab.
        stoi: A collections.defaultdict instance mapping token strings to
            numerical identifiers.
        itos: A list of token strings indexed by their numerical identifiers.
    """

    def __init__(self, counter, max_size=None, min_freq=1, specials=('<pad>',)):
        """Create a Vocab object from a collections.Counter.

        Arguments:
            counter: collections.Counter object holding the frequencies of
                each value found in the data.
            max_size: The maximum size of the vocabulary, or None for no
                maximum. Default: None.
            min_freq: The minimum frequency needed to include a token in the
                vocabulary. Values less than 1 will be set to 1. Default: 1.
            specials: The list of special tokens (e.g., padding or eos) that
                will be prepended to the vocabulary in addition to an <unk>
                token. Default: ['<pad>']
            vectors: One of either the available pretrained vectors
                or custom pretrained vectors (see Vocab.load_vectors);
                or a list of aforementioned vectors
        """
        self.freqs = counter
        counter = counter.copy()
        min_freq = max(min_freq, 1)
        counter.update(specials)

        self.stoi = defaultdict(_default_unk_index)
        self.stoi.update({tok: i for i, tok in enumerate(specials)})
        self.itos = list(specials)

        counter.subtract({tok: counter[tok] for tok in specials})
        max_size = None if max_size is None else max_size + len(self.itos)

        # sort by frequency, then alphabetically
        words_and_frequencies = sorted(counter.items(), key=lambda tup: tup[0])
        words_and_frequencies.sort(key=lambda tup: tup[1], reverse=True)

        for word, freq in words_and_frequencies:
            if freq < min_freq or len(self.itos) == max_size:
                break
            self.itos.append(word)
            self.stoi[word] = len(self.itos) - 1

    def __eq__(self, other):
        if self.freqs != other.freqs or self.stoi != other.stoi or self.itos != other.itos:
            return False
        return True

    def __len__(self):
        return len(self.itos)

    def extend(self, v, sort=False):
        words = sorted(v.itos) if sort else v.itos
        for w in words:
            if w not in self.stoi:
                self.itos.append(w)
                self.stoi[w] = len(self.itos) - 1

    @staticmethod
    def build_from_data(field_names, *args, unk_token=None, pad_token=None, init_token=None, eos_token=None, **kwargs):
        """Construct the Vocab object for this field from one or more datasets.

        Arguments:
            Positional arguments: Dataset objects or other iterable data
                sources from which to construct the Vocab object that
                represents the set of possible values for this field. If
                a Dataset object is provided, all columns corresponding
                to this field are used; individual columns can also be
                provided directly.
            Remaining keyword arguments: Passed to the constructor of Vocab.
        """
        counter = Counter()
        sources = []
        for arg in args:
            sources += [getattr(ex, name) for name in field_names for ex in arg]
        for data in sources:
            counter.update(data)
        specials = [unk_token, pad_token, init_token, eos_token]
        specials = [tok for tok in specials if tok is not None]
        return Vocab(counter, specials=specials, **kwargs)



#
# Copyright (c) 2018, Salesforce, Inc.
#                     The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import logging
import os
from pprint import pformat
from tqdm import tqdm

import torch

from . import models
from .data_utils.embeddings import load_embeddings
from .metrics import compute_metrics
from .tasks.registry import get_tasks
from .util import set_seed, preprocess_examples, load_config_json, make_data_loader, log_model_size, init_devices

logger = logging.getLogger(__name__)


def get_all_splits(args):
    splits = []
    for task in args.tasks:
        logger.info(f'Loading {task}')
        kwargs = {}
        if not 'train' in args.evaluate:
            kwargs['train'] = None
        if not 'valid' in args.evaluate:
            kwargs['validation'] = None
        if not 'test' in args.evaluate:
            kwargs['test'] = None

        kwargs['skip_cache'] = args.skip_cache
        kwargs['cached_path'] = os.path.join(args.cache, task.name)
        kwargs['subsample'] = args.subsample
        s = task.get_splits(root=args.data, lower=args.lower, **kwargs)[0]
        preprocess_examples(args, [task], [s], train=False)
        splits.append(s)
    return splits


def prepare_data(args, numericalizer, embeddings):
    splits = get_all_splits(args)
    logger.info(f'Vocabulary has {numericalizer.num_tokens} tokens from training')
    new_words = []
    for split in splits:
        new_words += numericalizer.grow_vocab(split)
        logger.info(f'Vocabulary has expanded to {numericalizer.num_tokens} tokens')

    for emb in embeddings:
        emb.grow_for_vocab(numericalizer.vocab, new_words)

    return splits


def run(args, numericalizer, val_sets, model, device):
    logger.info(f'Preparing iterators')
    if len(args.val_batch_size) == 1 and len(val_sets) > 1:
        args.val_batch_size *= len(val_sets)
    iters = [(name, make_data_loader(x, numericalizer, bs, device)) for name, x, bs in
             zip(args.tasks, val_sets, args.val_batch_size)]

    log_model_size(logger, model, args.model)
    model.to(device)

    decaScore = []
    model.eval()

    eval_dir = os.path.join(args.eval_dir, args.evaluate)
    os.makedirs(eval_dir, exist_ok=True)

    with torch.no_grad():
        for task, it in iters:
            logger.info(task.name)
            prediction_file_name = os.path.join(eval_dir, task.name + '.tsv')
            results_file_name = os.path.join(eval_dir, task.name + '.results.json')
            if os.path.exists(prediction_file_name):
                if args.overwrite:
                    logger.warning(f'{prediction_file_name} already exists -- overwriting **')
                else:
                    raise OSError(f'{prediction_file_name} already exists')
            if os.path.exists(results_file_name):
                if args.overwrite:
                    logger.warning(f'{results_file_name} already exists -- overwriting **')
                else:
                    raise OSError(f'{results_file_name} already exists')

            predictions = []
            answers = []
            with open(prediction_file_name, 'w' + ('' if args.overwrite else 'x')) as prediction_file:
                for batch_idx, batch in tqdm(enumerate(it), desc="Batches"):
                    _, batch_prediction = model(batch, iteration=1)

                    batch_prediction = numericalizer.reverse(batch_prediction, detokenize=task.detokenize,
                                                             field_name='answer')
                    predictions += batch_prediction
                    batch_answer = numericalizer.reverse(batch.answer.value.data, detokenize=task.detokenize,
                                                         field_name='answer')
                    answers += batch_answer

                    for i, example_prediction in enumerate(batch_prediction):
                        prediction_file.write(batch.example_id[i] + '\t' + example_prediction + '\n')

            if len(answers) > 0:
                metrics, answers = compute_metrics(predictions, answers, task.metrics, args=args)
                with open(results_file_name, 'w' + ('' if args.overwrite else '+')) as results_file:
                    results_file.write(json.dumps(metrics) + '\n')

                if not args.silent:
                    for i, (p, a) in enumerate(zip(predictions, answers)):
                        logger.info(f'Prediction {i + 1}: {p}\nAnswer {i + 1}: {a}\n')
                    logger.info(metrics)
                decaScore.append(metrics[task.metrics[0]])

    logger.info(f'Evaluated Tasks:\n')
    for i, (task, _) in enumerate(iters):
        logger.info(f'{task.name}: {decaScore[i]}')
    logger.info(f'-------------------')
    logger.info(f'DecaScore:  {sum(decaScore)}\n')
    logger.info(f'\nSummary: | {sum(decaScore)} | {" | ".join([str(x) for x in decaScore])} |\n')


def parse_argv(parser):
    parser.add_argument('--path', required=True)
    parser.add_argument('--evaluate', type=str, required=True, choices=['valid', 'test'], help='Which dataset to evaluate (test or dev)')
    parser.add_argument('--tasks',
                        default=['almond', 'squad', 'iwslt.en.de', 'cnn_dailymail', 'multinli.in.out', 'sst', 'srl',
                                 'zre', 'woz.en', 'wikisql', 'schema'], dest='task_names', nargs='+')
    parser.add_argument('--devices', default=[0], nargs='+', type=int,
                        help='a list of devices that can be used (multi-gpu currently WIP)')
    parser.add_argument('--seed', default=123, type=int, help='Random seed.')
    parser.add_argument('--data', default='.data/', type=str, help='where to load data from.')
    parser.add_argument('--embeddings', default='.embeddings/', type=str, help='where to save embeddings.')
    parser.add_argument('--checkpoint_name', default='best.pth',
                        help='Checkpoint file to use (relative to --path, defaults to best.pth)')
    parser.add_argument('--bleu', action='store_true', help='whether to use the bleu metric (always on for iwslt)')
    parser.add_argument('--rouge', action='store_true',
                        help='whether to use the bleu metric (always on for cnn, dailymail, and cnn_dailymail)')
    parser.add_argument('--overwrite', action='store_true', help='whether to overwrite previously written predictions')
    parser.add_argument('--silent', action='store_true', help='whether to print predictions to stdout')

    parser.add_argument('--skip_cache', action='store_true',
                        help='whether use exisiting cached splits or generate new ones')
    parser.add_argument('--eval_dir', type=str, required=True, help='use this directory to store eval results')
    parser.add_argument('--cache', default='.cache', type=str, help='where to save cached files')

    parser.add_argument('--saved_models', default='./saved_models', type=str,
                        help='directory where cached models should be loaded from')
    parser.add_argument('--subsample', default=20000000, type=int,
                        help='subsample the eval/test datasets (experimental)')


def main(args):
    load_config_json(args)
    set_seed(args)
    args.tasks = get_tasks(args.task_names, args)

    logger.info(f'Arguments:\n{pformat(vars(args))}')
    logger.info(f'Loading from {args.best_checkpoint}')

    devices = init_devices(args)
    save_dict = torch.load(args.best_checkpoint, map_location=devices[0])

    numericalizer, context_embeddings, question_embeddings, decoder_embeddings = \
        load_embeddings(args.embeddings, args.context_embeddings, args.question_embeddings, args.decoder_embeddings,
                        args.max_generative_vocab, logger)
    numericalizer.load(args.path)
    for emb in set(context_embeddings + question_embeddings + decoder_embeddings):
        emb.init_for_vocab(numericalizer.vocab)

    logger.info(f'Initializing Model')
    Model = getattr(models, args.model)
    model = Model(numericalizer, args, context_embeddings, question_embeddings, decoder_embeddings)
    model_dict = save_dict['model_state_dict']
    model.load_state_dict(model_dict)
    splits = prepare_data(args, numericalizer, set(context_embeddings + question_embeddings + decoder_embeddings))

    run(args, numericalizer, splits, model, devices[0])



#
# Copyright (c) 2020 The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from torch import nn

from .common import CombinedEmbedding, PackedLSTM


class BiLSTMEncoder(nn.Module):
    def __init__(self, numericalizer, args, context_embeddings, question_embeddings):
        super().__init__()
        self.args = args
        self.pad_idx = numericalizer.pad_id

        def dp(args):
            return args.dropout_ratio if args.rnn_layers > 1 else 0.

        self.context_embeddings = CombinedEmbedding(numericalizer, context_embeddings, args.rnn_dimension,
                                                    trained_dimension=args.trainable_encoder_embeddings,
                                                    project=True,
                                                    finetune_pretrained=args.train_context_embeddings)

        self.question_embeddings = CombinedEmbedding(numericalizer, question_embeddings, args.rnn_dimension,
                                                     trained_dimension=0,
                                                     project=True,
                                                     finetune_pretrained=args.train_question_embeddings)

        self.bilstm_context = PackedLSTM(args.rnn_dimension, args.rnn_dimension,
                                         batch_first=True, bidirectional=True, num_layers=args.rnn_layers,
                                         dropout=dp(args))

        self.bilstm_question = PackedLSTM(args.rnn_dimension, args.rnn_dimension,
                                          batch_first=True, bidirectional=True, num_layers=args.rnn_layers,
                                          dropout=dp(args))

    def set_train_context_embeddings(self, trainable):
        self.context_embeddings.set_trainable(trainable)

    def set_train_question_embeddings(self, trainable):
        self.question_embeddings.set_trainable(trainable)

    def forward(self, batch):
        context, context_lengths = batch.context.value, batch.context.length
        question, question_lengths = batch.question.value, batch.question.length

        context_padding = context.data == self.pad_idx
        question_padding = question.data == self.pad_idx

        context_embedded = self.context_embeddings(context, padding=context_padding)
        question_embedded = self.question_embeddings(question, padding=question_padding)

        # pick the top-most N transformer layers to pass to the decoder for cross-attention
        # (add 1 to account for the embedding layer - the decoder will drop it later)
        self_attended_context = context_embedded.all_layers[-(self.args.transformer_layers + 1):]
        final_context = context_embedded.last_layer
        final_question = question_embedded.last_layer

        final_context, context_rnn_state = self.bilstm_context(final_context, context_lengths)
        context_rnn_state = tuple(self.reshape_rnn_state(x) for x in context_rnn_state)

        final_question, question_rnn_state = self.bilstm_question(final_question, question_lengths)
        question_rnn_state = tuple(self.reshape_rnn_state(x) for x in question_rnn_state)

        return self_attended_context, final_context, context_rnn_state, final_question, question_rnn_state

    def reshape_rnn_state(self, h):
        # h is (num_layers * num_directions, batch, hidden_size)
        # we reshape to (num_layers, num_directions, batch, hidden_size)
        # transpose to (num_layers, batch, num_directions, hidden_size)
        # reshape to (num_layers, batch, num_directions * hidden_size)
        # also note that hidden_size is half the value of args.dimension

        return h.view(h.size(0) // 2, 2, h.size(1), h.size(2)) \
            .transpose(1, 2).contiguous() \
            .view(h.size(0) // 2, h.size(1), h.size(2) * 2).contiguous()



#
# Copyright (c) 2018-2019, Salesforce, Inc.
#                          The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import torch
import os
from collections import defaultdict
import logging
from transformers import AutoTokenizer, AutoModel, AutoConfig
from typing import NamedTuple, List

from .numericalizer.simple import SimpleNumericalizer
from .numericalizer.transformer import BertNumericalizer, XLMRobertaNumericalizer
from . import word_vectors
from .almond_embeddings import AlmondEmbeddings
from .pretrained_lstm_lm import PretrainedLTSMLM

_logger = logging.getLogger(__name__)


class EmbeddingOutput(NamedTuple):
    all_layers: List[torch.Tensor]
    last_layer: torch.Tensor


class WordVectorEmbedding(torch.nn.Module):
    def __init__(self, vec_collection):
        super().__init__()
        self._vec_collection = vec_collection
        self.dim = vec_collection.dim
        self.num_layers = 0
        self.embedding = None

    def init_for_vocab(self, vocab):
        vectors = torch.empty(len(vocab), self.dim, device=torch.device('cpu'))
        for ti, token in enumerate(vocab.itos):
            vectors[ti] = self._vec_collection[token.strip()]

        # wrap in a list so it will not be saved by torch.save and it will not
        # be moved around by .to() and similar methods
        self.embedding = [torch.nn.Embedding(len(vocab.itos), self.dim)]
        self.embedding[0].weight.data = vectors

    def grow_for_vocab(self, vocab, new_words):
        if not new_words:
            return
        new_vectors = []
        for word in new_words:
            new_vector = self._vec_collection[word]

            # charNgram returns  a [1, D] tensor, while Glove returns a [D] tensor
            # normalize to [1, D] so we can concat along the second dimension
            # and later concat all vectors along the first
            new_vector = new_vector if new_vector.dim() > 1 else new_vector.unsqueeze(0)
            new_vectors.append(new_vector)

        if new_vectors:
            self.embedding[0].weight.data = torch.cat([self.embedding[0].weight.data.cpu()] + new_vectors, dim=0)

    def forward(self, input: torch.Tensor, padding=None):
        last_layer = self.embedding[0](input.cpu()).to(input.device)
        return EmbeddingOutput(all_layers=[last_layer], last_layer=last_layer)

    def to(self, *args, **kwargs):
        # ignore attempts to move the word embedding, which should stay on CPU
        kwargs['device'] = torch.device('cpu')
        return super().to(*args, **kwargs)

    def cuda(self, device=None):
        # ignore attempts to move the word embedding
        pass


class TransformerEmbedding(torch.nn.Module):
    def __init__(self, model):
        super().__init__()
        model.config.output_hidden_states = True
        self.dim = model.config.hidden_size
        self.num_layers = model.config.num_hidden_layers
        self.model = model

    def init_for_vocab(self, vocab):
        self.model.resize_token_embeddings(len(vocab))

    def grow_for_vocab(self, vocab, new_words):
        self.model.resize_token_embeddings(len(vocab))

    def forward(self, input: torch.Tensor, padding=None):
        last_hidden_state, _pooled, hidden_states = self.model(input, attention_mask=(~padding).to(dtype=torch.float))

        return EmbeddingOutput(all_layers=hidden_states, last_layer=last_hidden_state)


class PretrainedLMEmbedding(torch.nn.Module):
    def __init__(self, model_name, cachedir):
        super().__init__()
        # map to CPU first, we will be moved to the right place later
        pretrained_save_dict = torch.load(os.path.join(cachedir, model_name), map_location=torch.device('cpu'))

        self.itos = pretrained_save_dict['vocab']
        self.stoi = defaultdict(lambda: 0, {
            w: i for i, w in enumerate(self.pretrained_decoder_vocab_itos)
        })
        self.dim = pretrained_save_dict['settings']['nhid']
        self.num_layers = 1
        self.model = PretrainedLTSMLM(rnn_type=pretrained_save_dict['settings']['rnn_type'],
                                      ntoken=len(self.pretrained_decoder_vocab_itos),
                                      emsize=pretrained_save_dict['settings']['emsize'],
                                      nhid=pretrained_save_dict['settings']['nhid'],
                                      nlayers=pretrained_save_dict['settings']['nlayers'],
                                      dropout=0.0)
        self.model.load_state_dict(pretrained_save_dict['model'], strict=True)

        self.vocab_to_pretrained = None

    def init_for_vocab(self, vocab):
        self.vocab_to_pretrained = torch.empty(len(self.vocab), dtype=torch.int64)

        unk_id = self.stoi['<unk>']
        for ti, token in enumerate(vocab.itos):
            if token in self.pretrained_decoder_vocab_stoi:
                self.vocab_to_pretrained[ti] = self.stoi[token]
            else:
                self.vocab_to_pretrained[ti] = unk_id

    def grow_for_vocab(self, vocab, new_words):
        self.init_for_vocab(vocab)

    def forward(self, input: torch.Tensor, padding=None):
        pretrained_indices = torch.gather(self.vocab_to_pretrained, dim=0, index=input)
        rnn_output = self.model(pretrained_indices)
        return EmbeddingOutput(all_layers=[rnn_output], last_layer=rnn_output)


def _is_bert(embedding_name):
    return embedding_name.startswith('bert-')


def _is_xlmr(embedding_name):
    return embedding_name.startswith('xlm-roberta-')


def _name_to_vector(emb_name, cachedir):
    if emb_name == 'glove':
        return WordVectorEmbedding(word_vectors.GloVe(cache=cachedir))
    elif emb_name == 'small_glove':
        return WordVectorEmbedding(word_vectors.GloVe(cache=cachedir, name="6B", dim=50))
    elif emb_name == 'char':
        return WordVectorEmbedding(word_vectors.CharNGram(cache=cachedir))
    elif emb_name == 'almond_type':
        return AlmondEmbeddings()
    elif emb_name.startswith('fasttext/'):
        # FIXME this should use the fasttext library
        return WordVectorEmbedding(word_vectors.FastText(cache=cachedir, language=emb_name[len('fasttext/'):]))
    elif emb_name.startswith('pretrained_lstm/'):
        return PretrainedLMEmbedding(emb_name[len('pretrained_lstm/'):], cachedir=cachedir)
    else:
        raise ValueError(f'Unrecognized embedding name {emb_name}')


def get_embedding_type(emb_name):
    if '@' in emb_name:
        return emb_name.split('@')[0]
    else:
        return emb_name


def load_embeddings(cachedir, context_emb_names, question_emb_names, decoder_emb_names,
                    max_generative_vocab=50000, logger=_logger):
    logger.info(f'Getting pretrained word vectors and pretrained models')

    context_emb_names = context_emb_names.split('+')
    question_emb_names = question_emb_names.split('+')
    decoder_emb_names = decoder_emb_names.split('+')

    all_vectors = {}
    context_vectors = []
    question_vectors = []
    decoder_vectors = []

    numericalizer = None
    numericalizer_type = None
    for emb_name in context_emb_names:
        if not emb_name:
            continue
        if emb_name in all_vectors:
            context_vectors.append(all_vectors[emb_name])
            continue

        emb_type = get_embedding_type(emb_name)
        if _is_bert(emb_name) or _is_xlmr(emb_name):
            if numericalizer is not None and numericalizer_type != emb_type:
                raise ValueError('Cannot specify multiple Transformer embeddings')

            config = AutoConfig.from_pretrained(emb_type, cache_dir=cachedir)
            config.output_hidden_states = True
            if numericalizer is None:
                if _is_bert(emb_type):
                    numericalizer = BertNumericalizer(emb_type, config=config,
                                                      max_generative_vocab=max_generative_vocab,
                                                      cache=cachedir)
                elif _is_xlmr(emb_type):
                    numericalizer = XLMRobertaNumericalizer(emb_type, config=config,
                                                            max_generative_vocab=max_generative_vocab,
                                                            cache=cachedir)
                numericalizer_type = emb_type

            # load the tokenizer once to ensure all files are downloaded
            AutoTokenizer.from_pretrained(emb_type, cache_dir=cachedir)

            context_vectors.append(
                TransformerEmbedding(AutoModel.from_pretrained(emb_type, config=config, cache_dir=cachedir)))
        else:
            if numericalizer is not None:
                logger.warning('Combining Transformer embeddings with other pretrained embeddings is unlikely to work')
            vec = _name_to_vector(emb_type, cachedir)
            all_vectors[emb_name] = vec
            context_vectors.append(vec)

    for emb_name in question_emb_names:
        if not emb_name:
            continue
        if emb_name in all_vectors:
            question_vectors.append(all_vectors[emb_name])
            continue

        emb_type = get_embedding_type(emb_name)
        if _is_bert(emb_type) or _is_xlmr(emb_name):
            if numericalizer is not None and numericalizer_type != emb_type:
                raise ValueError('Cannot specify multiple Transformer embeddings')

            config = AutoConfig.from_pretrained(emb_type, cache_dir=cachedir)
            config.output_hidden_states = True
            if numericalizer is None:
                if _is_bert(emb_type):
                    numericalizer = BertNumericalizer(emb_type, config=config,
                                                      max_generative_vocab=max_generative_vocab,
                                                      cache=cachedir)
                elif _is_xlmr(emb_type):
                    numericalizer = XLMRobertaNumericalizer(emb_type, config=config,
                                                            max_generative_vocab=max_generative_vocab,
                                                            cache=cachedir)
                numericalizer_type = emb_type

            # load the tokenizer once to ensure all files are downloaded
            AutoTokenizer.from_pretrained(emb_type, cache_dir=cachedir)

            question_vectors.append(
                TransformerEmbedding(AutoModel.from_pretrained(emb_type, config=config, cache_dir=cachedir)))
        else:
            if numericalizer is not None:
                logger.warning('Combining Transformer embeddings with other pretrained embeddings is unlikely to work')
            vec = _name_to_vector(emb_type, cachedir)
            all_vectors[emb_name] = vec
            question_vectors.append(vec)

    for emb_name in decoder_emb_names:
        if not emb_name:
            continue
        emb_type = get_embedding_type(emb_name)
        if _is_bert(emb_name) or _is_xlmr(emb_name):
            raise ValueError('Transformer embeddings cannot be specified in the decoder')

        if emb_name in all_vectors:
            decoder_vectors.append(all_vectors[emb_name])
        else:
            vec = _name_to_vector(emb_type, cachedir)
            all_vectors[emb_name] = vec
            decoder_vectors.append(vec)

    if numericalizer is None:
        numericalizer = SimpleNumericalizer(max_generative_vocab=max_generative_vocab, pad_first=False)

    return numericalizer, context_vectors, question_vectors, decoder_vectors



#
# Copyright (c) 2018, Salesforce, Inc.
#                     The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import logging
import logging.handlers
import math
import os
import time
from copy import deepcopy
from functools import partial
from pprint import pformat

import numpy as np
import torch
from tensorboardX import SummaryWriter

from . import arguments
from . import models
from .data_utils.embeddings import load_embeddings
from .data_utils.example import Example
from .util import elapsed_time, set_seed, preprocess_examples, get_trainable_params, make_data_loader, log_model_size, \
    init_devices
from .model_utils.parallel_utils import NamedTupleCompatibleDataParallel
from .model_utils.saver import Saver
from .validate import validate


def initialize_logger(args):
    # set up file logger
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    handler = logging.handlers.RotatingFileHandler(os.path.join(args.log_dir, f'train.log'),
                                                   maxBytes=1024 * 1024 * 10, backupCount=1)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(name)s - %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    handler.setLevel(logging.DEBUG)
    logger.addHandler(handler)
    logger.propagate = False

    return logger


def prepare_data(args, logger):
    train_sets, val_sets, aux_sets, vocab_sets = [], [], [], []
    for task in args.train_tasks:
        logger.info(f'Loading {task.name}')
        kwargs = {'test': None}
        kwargs['subsample'] = args.subsample
        kwargs['validation'] = None
        if args.use_curriculum:
            kwargs['curriculum'] = True
        kwargs['skip_cache'] = args.skip_cache
        kwargs['cached_path'] = os.path.join(args.cache, task.name)

        logger.info(f'Adding {task.name} to training datasets')
        split = task.get_splits(args.data, lower=args.lower, **kwargs)
        if args.use_curriculum:
            assert len(split) == 2
            aux_sets.append(split[1])
            logger.info(f'{task.name} has {len(split[1])} auxiliary examples')
        else:
            assert len(split) == 1
        train_sets.append(split[0])
        logger.info(f'{task.name} has {len(split[0])} training examples')
        if args.vocab_tasks is not None and task.name in args.vocab_tasks:
            vocab_sets.extend(split)

    for task in args.val_tasks:
        logger.info(f'Loading {task.name}')
        kwargs = {'test': None}
        kwargs['subsample'] = args.subsample
        kwargs['train'] = None
        kwargs['skip_cache'] = args.skip_cache
        kwargs['cached_path'] = os.path.join(args.cache, task.name)

        logger.info(f'Adding {task.name} to validation datasets')
        split = task.get_splits(args.data, lower=args.lower, **kwargs)
        assert len(split) == 1
        logger.info(f'{task.name} has {len(split[0])} validation examples')
        val_sets.append(split[0])
        if args.vocab_tasks is not None and task.name in args.vocab_tasks:
            vocab_sets.extend(split)

    numericalizer, context_embeddings, question_embeddings, decoder_embeddings = \
        load_embeddings(args.embeddings,
                        args.context_embeddings,
                        args.question_embeddings,
                        args.decoder_embeddings,
                        args.max_generative_vocab,
                        logger)
    if args.load is not None:
        numericalizer.load(args.save)
    else:
        vocab_sets = (train_sets + val_sets) if len(vocab_sets) == 0 else vocab_sets
        logger.info(f'Building vocabulary')
        numericalizer.build_vocab(Example.vocab_fields, vocab_sets)
        numericalizer.save(args.save)

    logger.info(f'Initializing encoder and decoder embeddings')
    for vec in set(context_embeddings + question_embeddings + decoder_embeddings):
        vec.init_for_vocab(numericalizer.vocab)

    logger.info(f'Vocabulary has {numericalizer.num_tokens} tokens')
    logger.debug(f'The first 200 tokens:')
    logger.debug(numericalizer.vocab.itos[:200])

    if args.use_curriculum:
        logger.info('Preprocessing auxiliary data for curriculum')
        preprocess_examples(args, args.train_tasks, aux_sets, logger, train=True)
    logger.info('Preprocessing training data')
    preprocess_examples(args, args.train_tasks, train_sets, logger, train=True)
    logger.info('Preprocessing validation data')
    preprocess_examples(args, args.val_tasks, val_sets, logger, train=args.val_filter)

    return numericalizer, context_embeddings, question_embeddings, decoder_embeddings, train_sets, val_sets, aux_sets

accumulated_batch_lengths = 0

def train_step(model, batch, iteration, opt, devices, lr_scheduler=None, grad_clip=None, pretraining=False,
               train_context_embeddings_after=None, train_question_embeddings_after=None,
               gradient_accumulation_steps=1):
    # Since the batch size is different in each call to this function due to dynamic batching, we need to keep track of
    # the total batch size
    global accumulated_batch_lengths
    model.train()
    model.module.set_train_context_embeddings(train_context_embeddings_after is not None and
                                              iteration > train_context_embeddings_after)
    model.module.set_train_question_embeddings(train_question_embeddings_after is not None and
                                               iteration > train_question_embeddings_after)
    if (iteration) % gradient_accumulation_steps == 0:
        opt.zero_grad()
    loss, predictions = model(batch, iteration, pretraining=pretraining)
    if torch.isnan(loss).any():
        raise RuntimeError('Got NaN loss')
    if len(devices) > 1:
        loss = loss.mean()
    non_accumulated_loss = loss.item()
    loss = loss*len(batch[0])
    accumulated_batch_lengths += len(batch[0])

    loss.backward()
    grad_norm = None
    if (iteration+1) % gradient_accumulation_steps == 0:
        for p in model.parameters():
            if p.grad is None:
                continue
            # print('p.grad = ', p.grad)
            p.grad /= accumulated_batch_lengths
        accumulated_batch_lengths = 0
        if grad_clip > 0.0:
            grad_norm = torch.nn.utils.clip_grad_norm_(model.params, grad_clip)
        opt.step()
        if lr_scheduler is not None:
            lr_scheduler.step()

    return non_accumulated_loss, grad_norm


def update_fraction(args, task_iteration):
    if args.curriculum_strategy == 'linear':
        next_fraction = args.curriculum_rate * task_iteration
    elif args.curriculum_strategy == 'exp':
        next_fraction = args.curriculum_rate * np.exp(task_iteration)

    fraction = min(args.curriculum_max_frac, next_fraction)

    return fraction


def should_validate(iteration, val_every, resume, start_iteration):
    if val_every is None:
        return False
    return (iteration % val_every == 0) or (resume and iteration == start_iteration)


def should_save(iteration, save_every):
    if save_every is None:
        return False
    return iteration % save_every == 0


def should_log(iteration, log_every):
    if log_every is None:
        return False
    return iteration % log_every == 0


def do_validate(iteration, args, model, numericalizer, val_iters, *,
                train_task, round_progress, task_progress, writer, logger):
    deca_score = 0
    for val_task_idx, (val_task, val_iter) in enumerate(val_iters):
        val_loss, metric_dict = validate(val_task, val_iter, model, logger, numericalizer,
                                         iteration, num_print=args.num_print, args=args)
        if val_loss is not None:
            log_entry = f'{args.timestamp}:{elapsed_time(logger)}:iteration_{iteration}:{round_progress}train_{train_task.name}:{task_progress}val_{val_task.name}:val_loss{val_loss.item():.4f}:'
            writer.add_scalar(f'loss/{val_task.name}/val', val_loss.item(), iteration)
        else:
            log_entry = f'{args.timestamp}:{elapsed_time(logger)}:iteration_{iteration}:{round_progress}train_{train_task.name}:{task_progress}val_{val_task.name}:'

        metric_entry = ''
        for metric_key, metric_value in metric_dict.items():
            metric_entry += f'{metric_key}_{metric_value:.2f}:'
        metric_entry = metric_entry[:-1]

        deca_score += metric_dict[val_task.metrics[0]]

        # val log
        logger.info(log_entry + metric_entry)
        if writer is not None:
            for metric_key, metric_value in metric_dict.items():
                writer.add_scalar(f'{val_task.name}/{metric_key}/val', metric_value, iteration)
    if writer is not None:
        writer.add_scalar('deca/val', deca_score, iteration)
    logger.info(
        f'{args.timestamp}:{elapsed_time(logger)}:iteration_{iteration}:{round_progress}train_{train_task.name}:{task_progress}val_deca:deca_{deca_score:.2f}')

    return deca_score


def maybe_save(iteration, model, opt, deca_score, best_decascore, *,
               saver, logger, train_task, round_progress, task_progress, timestamp, log_dir):
    should_save_best = False
    if deca_score is not None and (best_decascore is None or best_decascore < deca_score):
        best_decascore = deca_score
        should_save_best = True

    # punch through the nn.DataParallel to access the real model, otherwise we won't be able
    # to load this model later
    model_state_dict = model.module.state_dict()
    model_state_dict = {k: v.cpu() for k, v in model_state_dict.items()}

    save_model_state_dict = {
        'model_state_dict': model_state_dict,
        'best_decascore': best_decascore
    }
    save_opt_state_dict = opt.state_dict()
    save_opt_state_dict.update({'start_iteration': iteration})

    saver.save(save_model_state_dict, save_opt_state_dict, global_step=iteration)
    if should_save_best:
        logger.info(
            f'{timestamp}:{elapsed_time(logger)}:iteration_{iteration}:{round_progress}train_{train_task.name}:{task_progress}found new best model')
        torch.save(save_model_state_dict, os.path.join(log_dir, 'best.pth'))
        torch.save(save_opt_state_dict, os.path.join(log_dir, 'best_optim.pth'))

    return best_decascore


def do_log_training_loss(iteration, loss, *,
                         lr_scheduler, grad_norm, lr_rate,
                         num_examples, len_contexts, len_answers,
                         logger, train_task, round_progress, task_progress,
                         timestamp, writer, log_prefix):
    avg_batch_size = f'avbatch_{num_examples:.0f}_{len_contexts:.0f}_{len_answers:.0f}:'
    logger.info(
        f'{timestamp}:{elapsed_time(logger)}:iteration_{iteration}:{round_progress}train_{train_task.name}:{task_progress}{avg_batch_size}{log_prefix}/loss_{loss:.4f}')

    if writer is not None:
        writer.add_scalar(f'{log_prefix}/loss/{train_task.name}', loss, iteration)

        if lr_scheduler is not None:
            writer.add_scalar(f'{log_prefix}/lr', lr_scheduler.get_last_lr(), iteration)
        else:
            writer.add_scalar(f'{log_prefix}/lr', lr_rate, iteration)
        if grad_norm is not None:
            writer.add_scalar(f'{log_prefix}/norm', grad_norm, iteration)


def np_coin(prob):
    return np.random.uniform() < prob


def get_next_batch(train_iter, aux_iters, *, task, task_idx, task_fraction, use_curriculum):
    if use_curriculum:
        aux_iter = aux_iters[task_idx][1]
        prob = np_coin(task_fraction[task])
        if prob == 'aux':
            return next(aux_iter)
        else:
            assert prob == 'train'
            batch = next(train_iter)

    else:
        batch = next(train_iter)

    return batch


def train(args, devices, model, opt, lr_scheduler, train_sets, train_iterations, numericalizer, *,
          log_every, val_every, save_every, rounds, val_sets, aux_sets, writer, logger, log_prefix,
          start_iteration=1, rnd=1, best_decascore, use_curriculum, pretraining):
    """main training function"""
    local_loss, num_examples, len_contexts, len_answers, iteration = 0, 0, 0, 0, start_iteration

    train_iter_deep = deepcopy(train_iterations)

    task_iteration = dict()
    task_done = dict()
    task_fraction = dict()

    for task in args.train_tasks:
        task_iteration[task] = 1
        task_done[task] = False
        task_fraction[task] = 0.0

    saver = Saver(args.log_dir, args.max_to_keep)
    epoch = 0

    logger.info(f'Preparing iterators')
    main_device = devices[0]
    train_iters = [(task, make_data_loader(x, numericalizer, tok, main_device, train=True))
                   for task, x, tok in zip(args.train_tasks, train_sets, args.train_batch_tokens)]
    train_iters = [(task, iter(train_iter)) for task, train_iter in train_iters]

    val_iters = [(task, make_data_loader(x, numericalizer, bs, main_device, train=False))
                 for task, x, bs in zip(args.val_tasks, val_sets, args.val_batch_size)]

    aux_iters = []
    if use_curriculum:
        aux_iters = [(name, make_data_loader(x, numericalizer, tok, main_device, train=True))
                     for name, x, tok in zip(args.train_tasks, aux_sets, args.train_batch_tokens)]
        aux_iters = [(task, iter(aux_iter)) for task, aux_iter in aux_iters]

    zero_loss = 0
    logger.info(f'Begin {log_prefix}')

    while not all(task_done.values()):
        # For some number of rounds, we 'jump start' some subset of the tasks
        # by training them and not others
        # once the specified number of rounds is completed,
        # switch to normal round robin training
        if rnd < args.jump_start:
            train_iterations = [0] * len(train_iterations)
            for j in range(args.n_jump_start): train_iterations[j] = 1
        else:
            train_iterations = train_iter_deep

        for task_idx, (task, train_iter) in enumerate(train_iters):
            task_iterations = train_iterations[task_idx] if train_iterations is not None else None
            if task_iterations == 0:
                continue
            if task_iterations is not None and task_iteration[task] > task_iterations:
                task_done[task] = True
                continue

            batch = get_next_batch(train_iter, aux_iters, task=task, task_idx=task_idx,
                                   task_fraction=task_fraction, use_curriculum=use_curriculum)

            if iteration < start_iteration:
                # skip this iteration (this is done to ensure iterators are at the same position when resuming)
                task_iteration[task] += 1
                iteration += 1
                return

            task_progress = f'{task_iteration[task]}/{task_iterations}:' if task_iterations is not None else ''
            round_progress = f'round_{rnd}:' if rounds else ''

            # validate
            if should_validate(iteration, val_every, resume=args.resume, start_iteration=start_iteration):
                deca_score = do_validate(iteration, args, model, numericalizer, val_iters,
                                         train_task=task, round_progress=round_progress,
                                         task_progress=task_progress, writer=writer, logger=logger)

                # saving
                if should_save(iteration, save_every):
                    best_decascore = maybe_save(iteration, model, opt, deca_score, best_decascore,
                                                saver=saver, logger=logger, train_task=task,
                                                round_progress=round_progress, task_progress=task_progress,
                                                timestamp=args.timestamp, log_dir=args.log_dir)

            # param update
            loss, grad_norm = train_step(model, batch, iteration, opt, devices, lr_scheduler=lr_scheduler,
                                         grad_clip=args.grad_clip, pretraining=pretraining,
                                         gradient_accumulation_steps=args.gradient_accumulation_steps,
                                         train_context_embeddings_after=args.train_context_embeddings_after if
                                                                        args.train_context_embeddings else None,
                                         train_question_embeddings_after=args.train_question_embeddings_after if
                                                                         args.train_question_embeddings else None)
            if loss is None:
                logger.info(
                    'Encountered NAN loss during training... Continue training ignoring the current batch')
                continue
            if loss < 1e-5:
                zero_loss += 1
                if zero_loss >= 100:
                    logger.info('Found loss less than 1e-5 for 100 steps, stopping.')
                    return
            else:
                zero_loss = 0

            # update curriculum fraction
            if args.use_curriculum:
                task_fraction[task] = update_fraction(args, task_iteration[task])

            # train metrics
            local_loss += loss

            # train logs
            num_examples += batch.context.value.size(0)
            len_contexts += batch.context.value.size(1)
            len_answers += batch.answer.value.size(1)

            if should_log(iteration, log_every):
                local_loss /= log_every
                num_examples /= log_every
                len_contexts /= log_every
                len_answers /= log_every
                do_log_training_loss(iteration, local_loss,
                                     lr_scheduler=lr_scheduler, grad_norm=grad_norm, lr_rate=args.lr_rate,
                                     num_examples=num_examples, len_contexts=len_contexts, len_answers=len_answers,
                                     logger=logger, writer=writer, train_task=task, round_progress=round_progress,
                                     task_progress=task_progress, timestamp=args.timestamp, log_prefix=log_prefix)
                num_examples = 0
                len_contexts = 0
                len_answers = 0
                local_loss = 0

            # book keeping
            task_iteration[task] += 1
            iteration += 1

        # book keeping
        epoch += 1
        rnd += 1

    logger.info(f'{log_prefix} is done after {epoch} epochs')


def init_model(args, numericalizer, context_embeddings, question_embeddings, decoder_embeddings, devices, logger,
               save_dict):
    model_name = args.model
    logger.info(f'Initializing {model_name}')
    Model = getattr(models, model_name)
    model = Model(numericalizer, args, context_embeddings, question_embeddings, decoder_embeddings)
    params = get_trainable_params(model)
    log_model_size(logger, model, model_name)

    if save_dict is not None:
        logger.info(f'Loading model from {os.path.join(args.save, args.load)}')
        save_dict = torch.load(os.path.join(args.save, args.load))
        model.load_state_dict(save_dict['model_state_dict'])

    model.to(devices[0])
    model = NamedTupleCompatibleDataParallel(model, device_ids=devices)
    model.params = params

    return model


def get_transformer_learning_rate(i, *, dimension, warmup):
    i += 1
    return 1. / math.sqrt(dimension) * min(1 / math.sqrt(i), i / (warmup * math.sqrt(warmup)))


def get_sgd_learning_rate(i, *, warmup):
    i += 1
    return min(math.sqrt(warmup) / math.sqrt(i), i / warmup)


def init_opt(args, model, logger):
    if args.optimizer == 'adam':
        if args.transformer_lr:
            opt = torch.optim.Adam(model.params, lr=args.transformer_lr_multiply, betas=(0.9, 0.98), eps=1e-9,
                                   weight_decay=args.weight_decay)
            lr_lambda = partial(get_transformer_learning_rate, dimension=args.dimension, warmup=args.warmup)
            scheduler = torch.optim.lr_scheduler.LambdaLR(opt, lr_lambda)
        else:
            opt = torch.optim.Adam(model.params, lr=args.lr_rate, betas=(args.beta0, 0.999),
                                   weight_decay=args.weight_decay)
            scheduler = None
    elif args.optimizer == 'radam':
        import radam
        if args.transformer_lr:
            logger.warning('--transformer_lr has no effect with RAdam optimizer, warmup is never applied')
        opt = radam.RAdam(model.params, lr=args.lr_rate, betas=(args.beta0, 0.999), weight_decay=args.weight_decay)
        scheduler = None
    else:
        assert args.optimizer == 'sgd'
        if args.transformer_lr:
            opt = torch.optim.SGD(model.params, lr=args.transformer_lr_multiply, weight_decay=args.weight_decay, )
            lr_lambda = partial(get_sgd_learning_rate, warmup=args.warmup)
            scheduler = torch.optim.lr_scheduler.LambdaLR(opt, lr_lambda)
        else:
            opt = torch.optim.SGD(model.params, lr=args.lr_rate, weight_decay=args.weight_decay, )
            scheduler = None

    return opt, scheduler


def main(args):
    args = arguments.post_parse(args)
    if args is None:
        return

    set_seed(args)
    devices = init_devices(args, args.devices)
    logger = initialize_logger(args)
    logger.info(f'Arguments:\n{pformat(vars(args))}')

    save_dict = None
    if args.load is not None:
        logger.info(f'Loading vocab from {os.path.join(args.save, args.load)}')
        save_dict = torch.load(os.path.join(args.save, args.load))
    numericalizer, context_embeddings, question_embeddings, decoder_embeddings, train_sets, val_sets, aux_sets = \
        prepare_data(args, logger)
    if (args.use_curriculum and aux_sets is None) or (not args.use_curriculum and len(aux_sets)):
        logging.error('sth unpleasant is happening with curriculum')

    logger.info(f'Processing')
    logger.start = time.time()

    model = init_model(args, numericalizer, context_embeddings, question_embeddings, decoder_embeddings,
                       devices, logger, save_dict)
    opt, lr_scheduler = init_opt(args, model, logger)
    start_iteration = 1


    if save_dict is not None and args.resume:
        logger.info(f'Resuming Training from {os.path.splitext(args.load)[0]}_optim.pth')
        opt_state_dict = torch.load(os.path.join(args.save, f'{os.path.splitext(args.load)[0]}_optim.pth'))
        start_iteration = opt_state_dict.pop('start_iteration')
        logger.info(f'Starting iteration is {start_iteration}')
        opt.load_state_dict(opt_state_dict)

    if hasattr(args, 'tensorboard') and args.tensorboard:
        logger.info(f'Initializing Writer')
        writer = SummaryWriter(log_dir=args.tensorboard_dir, purge_step=start_iteration)
    else:
        writer = None

    if not args.resume and args.pretrain_context > 0:
        pretrain_opt, pretrain_lr_scheduler = init_opt(args, model, logger)
        train_iterations = [args.pretrain_context for _ in args.train_tasks]
        train(args, devices, model, pretrain_opt, pretrain_lr_scheduler, train_sets,
              train_iterations, numericalizer, val_sets=[], aux_sets=[], logger=logger, writer=writer,
              log_every=args.log_every, val_every=None, save_every=None, use_curriculum=False,
              rounds=len(train_sets) > 1, start_iteration=start_iteration, best_decascore=0,
              pretraining=True, log_prefix='pretrain')

    train(args, devices, model, opt, lr_scheduler, train_sets,
          args.train_iterations, numericalizer, val_sets=val_sets, aux_sets=aux_sets, logger=logger, writer=writer,
          log_every=args.log_every, val_every=args.val_every, save_every=args.save_every,
          rounds=len(train_sets) > 1, start_iteration=start_iteration, use_curriculum=args.use_curriculum,
          best_decascore=save_dict.get('best_decascore') if save_dict is not None else None,
          pretraining=False, log_prefix='training')



#
# Copyright (c) 2020 The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import torch
from torch import nn

from .common import CombinedEmbedding, LayerNorm, LinearFeedforward


class IdentityEncoder(nn.Module):
    def __init__(self, numericalizer, args, context_embeddings, question_embeddings):
        super().__init__()
        self.args = args
        self.pad_idx = numericalizer.pad_id

        self.encoder_embeddings = CombinedEmbedding(numericalizer, context_embeddings, args.dimension,
                                                    trained_dimension=0,
                                                    project=False,
                                                    finetune_pretrained=args.train_context_embeddings)

        if self.args.rnn_layers > 0 and self.args.rnn_dimension != self.args.dimension:
            self.dropout = nn.Dropout(args.dropout_ratio)
            self.projection = nn.Linear(self.encoder_embeddings.dimension, self.args.rnn_dimension, bias=False)
        else:
            self.dropout = None
            self.projection = None

        if self.args.rnn_layers > 0 and self.args.rnn_zero_state == 'average':
            self.pool = LinearFeedforward(args.dimension, args.dimension, 2 * args.rnn_dimension * args.rnn_layers,
                                          dropout=args.dropout_ratio)
            self.norm = LayerNorm(2 * args.rnn_dimension * args.rnn_layers)
        else:
            self.pool = None
            self.norm = None

    def set_train_context_embeddings(self, trainable):
        self.encoder_embeddings.set_trainable(trainable)

    def set_train_question_embeddings(self, trainable):
        pass

    def forward(self, batch):
        context, context_lengths = batch.context.value, batch.context.length
        question, question_lengths = batch.question.value, batch.question.length

        context_padding = context.data == self.pad_idx
        question_padding = question.data == self.pad_idx

        context_embedded = self.encoder_embeddings(context, padding=context_padding)
        question_embedded = self.encoder_embeddings(question, padding=question_padding)

        # pick the top-most N transformer layers to pass to the decoder for cross-attention
        # (add 1 to account for the embedding layer - the decoder will drop it later)
        self_attended_context = context_embedded.all_layers[-(self.args.transformer_layers + 1):]
        final_context = context_embedded.last_layer
        final_question = question_embedded.last_layer

        if self.projection is not None:
            final_context = self.dropout(final_context)
            final_context = self.projection(final_context)

            final_question = self.dropout(final_question)
            final_question = self.projection(final_question)

        context_rnn_state = None
        question_rnn_state = None
        if self.args.rnn_layers > 0:
            if self.args.rnn_zero_state == 'zero':
                batch_size = context.size(0)

                zero = torch.zeros(self.args.rnn_layers, batch_size, self.args.rnn_dimension,
                                   dtype=torch.float, requires_grad=False, device=context.device)
                context_rnn_state = (zero, zero)
                question_rnn_state = (zero, zero)
            else:
                assert self.args.rnn_zero_state == 'average'
                batch_size = context.size(0)

                masked_final_context = context_embedded.last_layer.masked_fill(context_padding.unsqueeze(2), 0)
                summed_context = torch.sum(masked_final_context, dim=1)
                average_context = summed_context / context_lengths.unsqueeze(1)

                packed_rnn_state = self.norm(self.pool(average_context))

                # packed_rnn_state is (batch, 2 * rnn_layers * rnn_dim)
                packed_rnn_state = packed_rnn_state.reshape(batch_size, 2, self.args.rnn_layers,
                                                            self.args.rnn_dimension)
                # transpose to (2, batch, rnn_layers, rnn_dimension)
                packed_rnn_state = packed_rnn_state.transpose(0, 1)
                # transpose to (2, rnn_layers, batch, rnn_dimension)
                packed_rnn_state = packed_rnn_state.transpose(1, 2)
                # convert to a tuple of two (rnn_layers, batch, rnn_dimension) tensors
                packed_rnn_state = packed_rnn_state.chunk(2, dim=0)
                context_rnn_state = (packed_rnn_state[0].squeeze(0), packed_rnn_state[1].squeeze(0))

        return self_attended_context, final_context, context_rnn_state, final_question, question_rnn_state



#
# Copyright (c) 2018, Salesforce, Inc.
#                     The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import torch
import random

from .example import Batch


class Iterator(torch.utils.data.IterableDataset):
    def __init__(self,
                 dataset: torch.utils.data.Dataset,
                 batch_size,
                 shuffle=False,
                 repeat=False,
                 batch_size_fn=None,
                 bucket_by_sort_key=False):
        self.dataset = dataset

        self.batch_size = batch_size
        self.shuffle = shuffle
        self.repeat = repeat

        if batch_size_fn is None:
            def batch_size_fn(new, count, sofar):
                return count
        self.batch_size_fn = batch_size_fn
        self.bucket_by_sort_key = bucket_by_sort_key

    def __len__(self):
        if self.repeat:
            raise NotImplementedError()
        else:
            return len(self.dataset)

    def __iter__(self) -> Batch:
        while True:
            if self.shuffle:
                dataset = list(self.dataset)
                random.shuffle(dataset)
            else:
                dataset = self.dataset

            if self.bucket_by_sort_key:
                batches = self._pool(dataset)
            else:
                batches = self._batch(dataset, self.batch_size)

            for minibatch in batches:
                yield minibatch

            if not self.repeat:
                break

    def _batch(self, data, batch_size):
        """Yield elements from data in chunks of batch_size."""
        minibatch = []
        size_so_far = 0
        for ex in data:
            minibatch.append(ex)
            size_so_far = self.batch_size_fn(ex, len(minibatch), size_so_far)
            if size_so_far == batch_size:
                yield minibatch
                minibatch, size_so_far = [], 0
            elif size_so_far > batch_size:
                if len(minibatch) == 1:  # if we only have one really big example
                    yield minibatch
                    minibatch, size_so_far = [], 0
                else:
                    yield minibatch[:-1]
                    minibatch, size_so_far = minibatch[-1:], self.batch_size_fn(ex, 1, 0)
                    if size_so_far > batch_size:  # if we add a really big example that needs to be on its own to a batch
                        yield minibatch
                        minibatch, size_so_far = [], 0
        if minibatch:
            yield minibatch

    def _pool(self, data):
        """Sort within buckets, then batch, then shuffle batches.

        Partitions data into chunks of size 100*batch_size, sorts examples within
        each chunk using sort_key, then batch these examples and shuffle the
        batches.
        """
        for p in self._batch(data, self.batch_size * 100):
            p_batch = self._batch(sorted(p, key=self.dataset.sort_key), self.batch_size)
            if self.shuffle:
                p_batch = list(p_batch)
                random.shuffle(p_batch)
            for b in p_batch:
                yield b
